@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -29,6 +28,15 @@
 #include <fluent-bit/flb_net_dns.h>
 #include <ares.h>
 
+#define FLB_NETWORK_DEFAULT_BACKLOG_SIZE              128
+#define FLB_NETWORK_UNIX_SOCKET_PEER_ADDRESS_TEMPLATE "pid_%s"
+#define FLB_NETWORK_ADDRESS_UNAVAILABLE               "unavailable"
+
+/* FLB_NETWORK_MAX_UNIX_ADDRESS_LENGTH should be enough for the
+ * string "PID " plus the string form of a signed 32 bit integer
+ * and a NULL character.
+ */
+
 /* Network connection setup */
 struct flb_net_setup {
     /* enable/disable keepalive support */
@@ -40,6 +48,17 @@ struct flb_net_setup {
     /* max time in seconds to wait for a established connection */
     int connect_timeout;
 
+    /* max time in seconds an incoming connection can take including the
+     * TLS handshake
+     */
+    int accept_timeout;
+
+    /* accept timeout log error (default: true) */
+    int accept_timeout_log_error;
+
+    /* max time in seconds to wait for blocking io calls */
+    int io_timeout;
+
     /* connect timeout log error (default: true) */
     int connect_timeout_log_error;
 
@@ -57,16 +76,19 @@ struct flb_net_setup {
 
     /* prioritize ipv4 results when trying to establish a connection*/
     int   dns_prefer_ipv4;
+
+    /* maximum number of allowed active TCP connections */
+    int max_worker_connections;
 };
 
 /* Defines a host service and it properties */
 struct flb_net_host {
-    int ipv6;              /* IPv6 required ?      */
-    flb_sds_t address;     /* Original address     */
-    int port;              /* TCP port             */
-    flb_sds_t name;        /* Hostname             */
-    flb_sds_t listen;      /* Listen interface     */
-    struct flb_uri *uri;   /* Extra URI parameters */
+    int ipv6;              /* IPv6 required ?        */
+    flb_sds_t address;     /* Original address       */
+    int port;              /* TCP port               */
+    flb_sds_t name;        /* Hostname               */
+    flb_sds_t listen;      /* Listen interface       */
+    struct flb_uri *uri;   /* Extra URI parameters   */
 };
 
 /* Defines an async DNS lookup context */
@@ -124,6 +146,7 @@ int flb_net_socket_reset(flb_sockfd_t fd);
 int flb_net_socket_tcp_nodelay(flb_sockfd_t fd);
 int flb_net_socket_blocking(flb_sockfd_t fd);
 int flb_net_socket_nonblocking(flb_sockfd_t fd);
+int flb_net_socket_rcv_buffer(flb_sockfd_t fd, int rcvbuf);
 int flb_net_socket_tcp_fastopen(flb_sockfd_t sockfd);
 
 /* Socket handling */
@@ -133,7 +156,7 @@ flb_sockfd_t flb_net_tcp_connect(const char *host, unsigned long port,
                                  char *source_addr, int connect_timeout,
                                  int is_async,
                                  void *async_ctx,
-                                 struct flb_upstream_conn *u_conn);
+                                 struct flb_connection *u_conn);
 
 flb_sockfd_t flb_net_udp_connect(const char *host, unsigned long port,
                                  char *source_addr);
@@ -141,13 +164,49 @@ flb_sockfd_t flb_net_udp_connect(const char *host, unsigned long port,
 int flb_net_tcp_fd_connect(flb_sockfd_t fd, const char *host, unsigned long port);
 flb_sockfd_t flb_net_server(const char *port, const char *listen_addr);
 flb_sockfd_t flb_net_server_udp(const char *port, const char *listen_addr);
+flb_sockfd_t flb_net_server_unix(const char *listen_path, int stream_mode,
+                                 int backlog);
 int flb_net_bind(flb_sockfd_t fd, const struct sockaddr *addr,
                  socklen_t addrlen, int backlog);
 int flb_net_bind_udp(flb_sockfd_t fd, const struct sockaddr *addr,
                  socklen_t addrlen);
 flb_sockfd_t flb_net_accept(flb_sockfd_t server_fd);
+
 int flb_net_address_to_str(int family, const struct sockaddr *addr,
                            char *output_buffer, size_t output_buffer_size);
-int flb_net_socket_ip_str(flb_sockfd_t fd, char **buf, int size, unsigned long *len);
+
+int flb_net_socket_peer_address(flb_sockfd_t fd,
+                                struct sockaddr_storage *output_buffer);
+
+int flb_net_socket_address_info(flb_sockfd_t fd,
+                                struct sockaddr_storage *address,
+                                unsigned short int *port_output_buffer,
+                                char *str_output_buffer,
+                                int str_output_buffer_size,
+                                size_t *str_output_data_size);
+
+int flb_net_socket_peer_ip_str(flb_sockfd_t fd,
+                               char *output_buffer,
+                               int output_buffer_size,
+                               size_t *output_data_size,
+                               int *output_address_family);
+
+int flb_net_socket_peer_ip_raw(flb_sockfd_t fd,
+                               char *output_buffer,
+                               int output_buffer_size,
+                               size_t *output_data_size,
+                               int *output_address_family);
+
+int flb_net_socket_peer_port(flb_sockfd_t fd,
+                             unsigned short int *output_buffer);
+
+int flb_net_socket_peer_info(flb_sockfd_t fd,
+                             unsigned short int *port_output_buffer,
+                             struct sockaddr_storage *raw_output_buffer,
+                             char *str_output_buffer,
+                             int str_output_buffer_size,
+                             size_t *str_output_data_size);
+
+size_t flb_network_address_size(struct sockaddr_storage *address);
 
 #endif
