@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -29,6 +28,8 @@
 #include <fluent-bit/flb_filter.h>
 #include <fluent-bit/flb_output.h>
 
+#include <fluent-bit/flb_hash.h>
+
 struct calyptia {
     /* config map options */
     flb_sds_t api_key;
@@ -37,6 +38,11 @@ struct calyptia {
     flb_sds_t cloud_port;
     flb_sds_t machine_id;
 
+/* used for reporting chunk trace records. */
+#ifdef FLB_HAVE_CHUNK_TRACE
+    flb_sds_t pipeline_id;
+#endif /* FLB_HAVE_CHUNK_TRACE */
+
     int cloud_tls;
     int cloud_tls_verify;
 
@@ -46,7 +52,15 @@ struct calyptia {
     /* instances */
     struct flb_input_instance *i;
     struct flb_output_instance *o;
+    struct flb_input_instance *fleet;
     struct flb_custom_instance *ins;
+
+    /* Fleet configuration */
+    flb_sds_t fleet_id;                   /* fleet-id  */
+    flb_sds_t fleet_name;
+    flb_sds_t fleet_config_dir;           /* fleet configuration directory */
+    int fleet_interval_sec;
+    int fleet_interval_nsec;
 };
 
 /*
@@ -88,16 +102,18 @@ static void pipeline_config_add_properties(flb_sds_t *buf, struct mk_list *props
     mk_list_foreach(head, props) {
         kv = mk_list_entry(head, struct flb_kv, _head);
 
-        flb_sds_printf(buf, "    %s ", kv->key);
+        if (kv->key != NULL && kv->val != NULL) {
+            flb_sds_printf(buf, "    %s ", kv->key);
 
-        if (is_sensitive_property(kv->key)) {
-            flb_sds_printf(buf, "--redacted--");
-        }
-        else {
-            flb_sds_printf(buf, kv->val);
-        }
+            if (is_sensitive_property(kv->key)) {
+                flb_sds_cat_safe(buf, "--redacted--", strlen("--redacted--"));
+            }
+            else {
+                flb_sds_cat_safe(buf, kv->val, strlen(kv->val));
+            }
 
-        flb_sds_printf(buf, "\n");
+            flb_sds_cat_safe(buf, "\n", 1);
+        }
     }
 }
 
@@ -111,6 +127,7 @@ flb_sds_t custom_calyptia_pipeline_config_get(struct flb_config *ctx)
     struct flb_output_instance *o_ins;
 
     buf = flb_sds_create_size(2048);
+
     if (!buf) {
         return NULL;
     }
@@ -120,9 +137,11 @@ flb_sds_t custom_calyptia_pipeline_config_get(struct flb_config *ctx)
         i_ins = mk_list_entry(head, struct flb_input_instance, _head);
         flb_sds_printf(&buf, "[INPUT]\n");
         flb_sds_printf(&buf, "    name %s\n", i_ins->name);
+
         if (i_ins->alias) {
             flb_sds_printf(&buf, "    alias %s\n", i_ins->alias);
         }
+
         if (i_ins->tag) {
             flb_sds_printf(&buf, "    tag %s\n", i_ins->tag);
         }
@@ -140,6 +159,7 @@ flb_sds_t custom_calyptia_pipeline_config_get(struct flb_config *ctx)
     /* Config: [FILTER] */
     mk_list_foreach(head, &ctx->filters) {
         f_ins = mk_list_entry(head, struct flb_filter_instance, _head);
+
         flb_sds_printf(&buf, "[FILTER]\n");
         flb_sds_printf(&buf, "    name  %s\n", f_ins->name);
         flb_sds_printf(&buf, "    match %s\n", f_ins->match);
@@ -151,8 +171,10 @@ flb_sds_t custom_calyptia_pipeline_config_get(struct flb_config *ctx)
     /* Config: [OUTPUT] */
     mk_list_foreach(head, &ctx->outputs) {
         o_ins = mk_list_entry(head, struct flb_output_instance, _head);
+
         flb_sds_printf(&buf, "[OUTPUT]\n");
         flb_sds_printf(&buf, "    name  %s\n", o_ins->name);
+
         if (o_ins->match) {
             flb_sds_printf(&buf, "    match %s\n", o_ins->match);
         }
@@ -186,6 +208,7 @@ flb_sds_t custom_calyptia_pipeline_config_get(struct flb_config *ctx)
             }
         }
 #endif
+
         if (o_ins->retry_limit == FLB_OUT_RETRY_UNLIMITED) {
             flb_sds_printf(&buf, "    retry_limit no_limits\n");
         }
@@ -207,20 +230,176 @@ flb_sds_t custom_calyptia_pipeline_config_get(struct flb_config *ctx)
     return buf;
 }
 
+static struct flb_output_instance *setup_cloud_output(struct flb_config *config, struct calyptia *ctx)
+{
+    int ret;
+    struct flb_output_instance *cloud;
+    struct mk_list *head;
+    struct flb_slist_entry *key = NULL;
+    struct flb_slist_entry *val = NULL;
+    flb_sds_t label;
+    struct flb_config_map_val *mv;
+
+    cloud = flb_output_new(config, "calyptia", ctx, FLB_FALSE);
+
+    if (!cloud) {
+        flb_plg_error(ctx->ins, "could not load Calyptia Cloud connector");
+        flb_free(ctx);
+        return NULL;
+    }
+
+    /* direct connect / routing */
+    ret = flb_router_connect_direct(ctx->i, cloud);
+
+    if (ret != 0) {
+        flb_plg_error(ctx->ins, "could not load Calyptia Cloud connector");
+        flb_free(ctx);
+        return NULL;
+    }
+
+    if (ctx->add_labels && mk_list_size(ctx->add_labels) > 0) {
+
+        /* iterate all 'add_label' definitions */
+        flb_config_map_foreach(head, mv, ctx->add_labels) {
+            key = mk_list_entry_first(mv->val.list, struct flb_slist_entry, _head);
+            val = mk_list_entry_last(mv->val.list, struct flb_slist_entry, _head);
+
+            label = flb_sds_create_size(strlen(key->str) + strlen(val->str) + 1);
+
+            if (!label) {
+                flb_free(ctx);
+                return NULL;
+            }
+
+            flb_sds_printf(&label, "%s %s", key->str, val->str);
+            flb_output_set_property(cloud, "add_label", label);
+            flb_sds_destroy(label);
+        }
+    }
+
+    flb_output_set_property(cloud, "match", "_calyptia_cloud");
+    flb_output_set_property(cloud, "api_key", ctx->api_key);
+
+    if (ctx->store_path) {
+        flb_output_set_property(cloud, "store_path", ctx->store_path);
+    }
+
+    if (ctx->machine_id) {
+        flb_output_set_property(cloud, "machine_id", ctx->machine_id);
+    }
+
+    /* Override network details: development purposes only */
+    if (ctx->cloud_host) {
+        flb_output_set_property(cloud, "cloud_host", ctx->cloud_host);
+    }
+
+    if (ctx->cloud_port) {
+        flb_output_set_property(cloud, "cloud_port", ctx->cloud_port);
+    }
+
+    if (ctx->cloud_tls) {
+        flb_output_set_property(cloud, "tls", "true");
+    }
+    else {
+        flb_output_set_property(cloud, "tls", "false");
+    }
+
+    if (ctx->cloud_tls_verify) {
+        flb_output_set_property(cloud, "tls.verify", "true");
+    }
+    else {
+        flb_output_set_property(cloud, "tls.verify", "false");
+    }
+
+    if (ctx->fleet_id) {
+        flb_output_set_property(cloud, "fleet_id", ctx->fleet_id);
+        label = flb_sds_create_size(strlen("fleet_id") + strlen(ctx->fleet_id) + 1);
+
+        if (!label) {
+            flb_free(ctx);
+            return NULL;
+        }
+
+        flb_sds_printf(&label, "fleet_id %s", ctx->fleet_id);
+        flb_output_set_property(cloud, "add_label", label);
+        flb_sds_destroy(label);
+    }
+
+#ifdef FLB_HAVE_CHUNK_TRACE
+    flb_output_set_property(cloud, "pipeline_id", ctx->pipeline_id);
+#endif /* FLB_HAVE_CHUNK_TRACE */
+
+    return cloud;
+}
+
+static flb_sds_t sha256_to_hex(unsigned char *sha256)
+{
+    int idx;
+    flb_sds_t hex;
+    flb_sds_t tmp;
+
+    hex = flb_sds_create_size(64);
+
+    if (!hex) {
+        return NULL;
+    }
+
+    for (idx = 0; idx < 32; idx++) {
+        tmp = flb_sds_printf(&hex, "%02x", sha256[idx]);
+
+        if (!tmp) {
+            flb_sds_destroy(hex);
+            return NULL;
+        }
+
+        hex = tmp;
+    }
+
+    flb_sds_len_set(hex, 64);
+    return hex;
+}
+
+static flb_sds_t get_machine_id(struct calyptia *ctx)
+{
+    int ret;
+    char *buf;
+    size_t blen;
+    unsigned char sha256_buf[64] = {0};
+
+    /* retrieve raw machine id */
+    ret = flb_utils_get_machine_id(&buf, &blen);
+
+    if (ret == -1) {
+        flb_plg_error(ctx->ins, "could not obtain machine id");
+        return NULL;
+    }
+
+    ret = flb_hash_simple(FLB_HASH_SHA256,
+                          (unsigned char *) buf,
+                          blen,
+                          sha256_buf,
+                          sizeof(sha256_buf));
+    flb_free(buf);
+
+    if (ret != FLB_CRYPTO_SUCCESS) {
+        return NULL;
+    }
+
+    /* convert to hex */
+    return sha256_to_hex(sha256_buf);
+}
+
 static int cb_calyptia_init(struct flb_custom_instance *ins,
                             struct flb_config *config,
                             void *data)
 {
     int ret;
     struct calyptia *ctx;
-    struct mk_list *head;
-    struct flb_config_map_val *mv;
-    struct flb_slist_entry *k = NULL;
-    struct flb_slist_entry *v = NULL;
+    int is_fleet_mode;
     (void) data;
-    flb_sds_t kv;
 
     ctx = flb_calloc(1, sizeof(struct calyptia));
+
     if (!ctx) {
         flb_errno();
         return -1;
@@ -229,6 +408,7 @@ static int cb_calyptia_init(struct flb_custom_instance *ins,
 
     /* Load the config map */
     ret = flb_custom_config_map_set(ins, (void *) ctx);
+
     if (ret == -1) {
         flb_free(ctx);
         return -1;
@@ -237,83 +417,94 @@ static int cb_calyptia_init(struct flb_custom_instance *ins,
     /* map instance and local context */
     flb_custom_set_context(ins, ctx);
 
+    /* If no machine_id has been provided via a configuration option get it from the local machine-id. */
+    if (!ctx->machine_id) {
+        /* machine id */
+        ctx->machine_id = get_machine_id(ctx);
+
+        if (ctx->machine_id == NULL) {
+            flb_plg_error(ctx->ins, "unable to retrieve machine_id");
+            return -1;
+        }
+    }
+
     /* input collector */
     ctx->i = flb_input_new(config, "fluentbit_metrics", NULL, FLB_TRUE);
+
     if (!ctx->i) {
         flb_plg_error(ctx->ins, "could not load metrics collector");
         return -1;
     }
+
     flb_input_set_property(ctx->i, "tag", "_calyptia_cloud");
     flb_input_set_property(ctx->i, "scrape_on_start", "true");
     flb_input_set_property(ctx->i, "scrape_interval", "30");
 
+    if (ctx->fleet_name || ctx->fleet_id) {
+        is_fleet_mode = FLB_TRUE;
+    }
+    else {
+        is_fleet_mode = FLB_FALSE;
+    }
+
     /* output cloud connector */
-    ctx->o = flb_output_new(config, "calyptia", ctx, FLB_FALSE);
-    if (!ctx->o) {
-        flb_plg_error(ctx->ins, "could not load Calyptia Cloud connector");
-        flb_free(ctx);
-        return -1;
-    }
+    if ((is_fleet_mode == FLB_TRUE && ctx->fleet_id != NULL) ||
+        (is_fleet_mode == FLB_FALSE)) {
+        ctx->o = setup_cloud_output(config, ctx);
 
-    /* direct connect / routing */
-    ret = flb_router_connect_direct(ctx->i, ctx->o);
-    if (ret != 0) {
-        flb_plg_error(ctx->ins, "could not load Calyptia Cloud connector");
-        flb_free(ctx);
-        return -1;
-    }
-
-    if (ctx->add_labels && mk_list_size(ctx->add_labels) > 0) {
-        /* iterate all 'add_label' definitions */
-        flb_config_map_foreach(head, mv, ctx->add_labels) {
-            k = mk_list_entry_first(mv->val.list, struct flb_slist_entry, _head);
-            v = mk_list_entry_last(mv->val.list, struct flb_slist_entry, _head);
-            kv = flb_sds_create_size(strlen(k->str) + strlen(v->str) + 1);
-            if(!kv) {
-                flb_free(ctx);
-                return -1;
-            }
-
-            flb_sds_printf(&kv, "%s %s", k->str, v->str);
-            flb_output_set_property(ctx->o, "add_label", kv);
-            flb_sds_destroy(kv);
+        if (ctx->o == NULL) {
+            return -1;
         }
     }
 
-    flb_output_set_property(ctx->o, "match", "_calyptia_cloud");
-    flb_output_set_property(ctx->o, "api_key", ctx->api_key);
-    if (ctx->store_path) {
-        flb_output_set_property(ctx->o, "store_path", ctx->store_path);
+    if (ctx->fleet_id || ctx->fleet_name) {
+        ctx->fleet =  flb_input_new(config, "calyptia_fleet", NULL, FLB_FALSE);
+
+        if (!ctx->fleet) {
+            flb_plg_error(ctx->ins, "could not load Calyptia Fleet plugin");
+            return -1;
+        }
+
+        if (ctx->fleet_name) {
+            // TODO: set this once the fleet_id has been retrieved...
+            // flb_output_set_property(ctx->o, "fleet_id", ctx->fleet_id);
+            flb_input_set_property(ctx->fleet, "fleet_name", ctx->fleet_name);            
+        }
+        else {
+            flb_output_set_property(ctx->o, "fleet_id", ctx->fleet_id);
+            flb_input_set_property(ctx->fleet, "fleet_id", ctx->fleet_id);
+        }
+
+        flb_input_set_property(ctx->fleet, "api_key", ctx->api_key);
+        flb_input_set_property(ctx->fleet, "host", ctx->cloud_host);
+        flb_input_set_property(ctx->fleet, "port", ctx->cloud_port);
+
+        if (ctx->cloud_tls == 1) {
+            flb_input_set_property(ctx->fleet, "tls", "on");
+        }
+        else {
+            flb_input_set_property(ctx->fleet, "tls", "off");
+        }
+
+        if (ctx->cloud_tls_verify == 1) {
+            flb_input_set_property(ctx->fleet, "tls.verify", "on");
+        }
+        else {
+            flb_input_set_property(ctx->fleet, "tls.verify", "off");
+        }
+
+        if (ctx->fleet_config_dir) {
+            flb_input_set_property(ctx->fleet, "config_dir", ctx->fleet_config_dir);
+        }
+
+        if (ctx->machine_id) {
+            flb_input_set_property(ctx->fleet, "machine_id", ctx->machine_id);
+        }
     }
 
-    if (ctx->machine_id) {
-        flb_output_set_property(ctx->o, "machine_id", ctx->machine_id);
+    if (ctx->o) {
+        flb_router_connect(ctx->i, ctx->o);
     }
-
-    /* Override network details: development purposes only */
-    if (ctx->cloud_host) {
-        flb_output_set_property(ctx->o, "cloud_host", ctx->cloud_host);
-    }
-
-    if (ctx->cloud_port) {
-        flb_output_set_property(ctx->o, "cloud_port", ctx->cloud_port);
-    }
-
-    if (ctx->cloud_tls) {
-        flb_output_set_property(ctx->o, "tls", "true");
-    }
-    else {
-        flb_output_set_property(ctx->o, "tls", "false");
-    }
-
-    if (ctx->cloud_tls_verify) {
-        flb_output_set_property(ctx->o, "tls.verify", "true");
-    }
-    else {
-        flb_output_set_property(ctx->o, "tls.verify", "false");
-    }
-
-    flb_router_connect(ctx->i, ctx->o);
     flb_plg_info(ins, "custom initialized!");
     return 0;
 }
@@ -344,13 +535,13 @@ static struct flb_config_map config_map[] = {
     },
 
     {
-     FLB_CONFIG_MAP_STR, "calyptia_host", NULL,
+     FLB_CONFIG_MAP_STR, "calyptia_host", "cloud-api.calyptia.com",
      0, FLB_TRUE, offsetof(struct calyptia, cloud_host),
      ""
     },
 
     {
-     FLB_CONFIG_MAP_STR, "calyptia_port", NULL,
+     FLB_CONFIG_MAP_STR, "calyptia_port", "443",
      0, FLB_TRUE, offsetof(struct calyptia, cloud_port),
      ""
     },
@@ -377,6 +568,39 @@ static struct flb_config_map config_map[] = {
      0, FLB_TRUE, offsetof(struct calyptia, machine_id),
      "Custom machine_id to be used when registering agent"
     },
+    {
+     FLB_CONFIG_MAP_STR, "fleet_id", NULL,
+     0, FLB_TRUE, offsetof(struct calyptia, fleet_id),
+     "Fleet id to be used when registering agent in a fleet"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "fleet.config_dir", NULL,
+     0, FLB_TRUE, offsetof(struct calyptia, fleet_config_dir),
+     "Base path for the configuration directory."
+    },
+    {
+      FLB_CONFIG_MAP_INT, "fleet.interval_sec", "-1",
+      0, FLB_TRUE, offsetof(struct calyptia, fleet_interval_sec),
+      "Set the collector interval"
+    },
+    {
+      FLB_CONFIG_MAP_INT, "fleet.interval_nsec", "-1",
+      0, FLB_TRUE, offsetof(struct calyptia, fleet_interval_nsec),
+      "Set the collector interval (nanoseconds)"
+    },
+    {
+     FLB_CONFIG_MAP_STR, "fleet_name", NULL,
+     0, FLB_TRUE, offsetof(struct calyptia, fleet_name),
+     "Fleet name to be used when registering agent in a fleet"
+    },
+
+#ifdef FLB_HAVE_CHUNK_TRACE
+    {
+     FLB_CONFIG_MAP_STR, "pipeline_id", NULL,
+     0, FLB_TRUE, offsetof(struct calyptia, pipeline_id),
+     "Pipeline ID for reporting to calyptia cloud."
+    },
+#endif /* FLB_HAVE_CHUNK_TRACE */
 
     /* EOF */
     {0}
