@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -21,6 +20,7 @@
 #include <fluent-bit/flb_output_plugin.h>
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_pack.h>
+#include <fluent-bit/flb_config_map.h>
 
 #include "datadog.h"
 #include "datadog_conf.h"
@@ -33,6 +33,7 @@ struct flb_out_datadog *flb_datadog_conf_create(struct flb_output_instance *ins,
     struct flb_upstream *upstream;
     const char *api_key;
     const char *tmp;
+    flb_sds_t tmp_sds;
 
     int ret;
     char *protocol = NULL;
@@ -49,6 +50,13 @@ struct flb_out_datadog *flb_datadog_conf_create(struct flb_output_instance *ins,
     ctx->ins = ins;
     ctx->nb_additional_entries = 0;
 
+    ret = flb_output_config_map_set(ins, (void *) ctx);
+    if (ret == -1) {
+        flb_plg_error(ins, "flb_output_config_map_set failed");
+        flb_free(ctx);
+        return NULL;
+    }
+
     tmp = flb_output_get_property("proxy", ins);
     if (tmp) {
         ret = flb_utils_url_split(tmp, &protocol, &host, &port, &uri);
@@ -60,7 +68,6 @@ struct flb_out_datadog *flb_datadog_conf_create(struct flb_output_instance *ins,
 
         ctx->proxy_host = host;
         ctx->proxy_port = atoi(port);
-        ctx->proxy = tmp;
         flb_free(protocol);
         flb_free(port);
         flb_free(uri);
@@ -69,90 +76,74 @@ struct flb_out_datadog *flb_datadog_conf_create(struct flb_output_instance *ins,
     /* use TLS ? */
     if (ins->use_tls == FLB_TRUE) {
         io_flags = FLB_IO_TLS;
-        ctx->scheme = flb_sds_create("https://");
+        tmp_sds = flb_sds_create("https://");
     }
     else {
         io_flags = FLB_IO_TCP;
-        ctx->scheme = flb_sds_create("http://");
+        tmp_sds = flb_sds_create("http://");
     }
+    if (!tmp_sds) {
+        flb_errno();
+        flb_datadog_conf_destroy(ctx);
+        return NULL;
+    }
+    ctx->scheme = tmp_sds;
     flb_plg_debug(ctx->ins, "scheme: %s", ctx->scheme);
 
     /* configure URI */
     api_key = flb_output_get_property("apikey", ins);
-    if (api_key) {
-        ctx->api_key = flb_sds_create(api_key);
-    }
-    else {
+    if (api_key == NULL) {
         flb_plg_error(ctx->ins, "no ApiKey configuration key defined");
         flb_datadog_conf_destroy(ctx);
         return NULL;
     }
 
-    /* Include Tag key */
-    tmp = flb_output_get_property("include_tag_key", ins);
-    if (tmp) {
-        ctx->include_tag_key = flb_utils_bool(tmp);
-    }
-    else {
-        ctx->include_tag_key = FLB_FALSE;
-    }
-
     /* Tag Key */
     if (ctx->include_tag_key == FLB_TRUE) {
         ctx->nb_additional_entries++;
-        tmp = flb_output_get_property("tag_key", ins);
-        if (tmp) {
-            ctx->tag_key = flb_sds_create(tmp);
-        }
-        else {
-            ctx->tag_key = flb_sds_create(FLB_DATADOG_DEFAULT_TAG_KEY);
-        }
     }
 
     tmp = flb_output_get_property("dd_source", ins);
     if (tmp) {
         ctx->nb_additional_entries++;
-        ctx->dd_source = flb_sds_create(tmp);
     }
 
     tmp = flb_output_get_property("dd_service", ins);
     if (tmp) {
         ctx->nb_additional_entries++;
-        ctx->dd_service = flb_sds_create(tmp);
     }
 
     tmp = flb_output_get_property("dd_tags", ins);
     if (tmp) {
         ctx->nb_additional_entries++;
-        ctx->dd_tags = flb_sds_create(tmp);
-    }
-
-    tmp = flb_output_get_property("dd_message_key", ins);
-    if (tmp) {
-        ctx->dd_message_key = flb_sds_create(tmp);
     }
 
     tmp = flb_output_get_property("provider", ins);
     ctx->remap = tmp && (strlen(tmp) == strlen(FLB_DATADOG_REMAP_PROVIDER)) && \
         (strncmp(tmp, FLB_DATADOG_REMAP_PROVIDER, strlen(tmp)) == 0);
 
-    ctx->uri = flb_sds_create("/v1/input/");
+    ctx->uri = flb_sds_create("/api/v2/logs");
     if (!ctx->uri) {
         flb_plg_error(ctx->ins, "error on uri generation");
         flb_datadog_conf_destroy(ctx);
         return NULL;
     }
-    /* Add the api_key to the URI */
-    ctx->uri = flb_sds_cat(ctx->uri, ctx->api_key, flb_sds_len(ctx->api_key));
+
     flb_plg_debug(ctx->ins, "uri: %s", ctx->uri);
 
     /* Get network configuration */
     if (!ins->host.name) {
-        ctx->host = flb_sds_create(FLB_DATADOG_DEFAULT_HOST);
+        tmp_sds = flb_sds_create(FLB_DATADOG_DEFAULT_HOST);
     }
     else {
-        ctx->host = flb_sds_create(ins->host.name);
+        tmp_sds = flb_sds_create(ins->host.name);
     }
+    if (!tmp_sds) {
+        flb_errno();
+        flb_datadog_conf_destroy(ctx);
+        return NULL;
+    }
+    ctx->host = tmp_sds;
     flb_plg_debug(ctx->ins, "host: %s", ctx->host);
 
     if (ins->host.port != 0) {
@@ -167,13 +158,6 @@ struct flb_out_datadog *flb_datadog_conf_create(struct flb_output_instance *ins,
     flb_plg_debug(ctx->ins, "port: %i", ctx->port);
 
     /* Date tag for JSON output */
-    tmp = flb_output_get_property("json_date_key", ins);
-    if (tmp) {
-        ctx->json_date_key = flb_sds_create(tmp);
-    }
-    else {
-        ctx->json_date_key = flb_sds_create(FLB_DATADOG_DEFAULT_TIME_KEY);
-    }
     ctx->nb_additional_entries++;
     flb_plg_debug(ctx->ins, "json_date_key: %s", ctx->json_date_key);
 
@@ -230,27 +214,6 @@ int flb_datadog_conf_destroy(struct flb_out_datadog *ctx)
     if (ctx->uri) {
         flb_sds_destroy(ctx->uri);
     }
-    if (ctx->api_key) {
-        flb_sds_destroy(ctx->api_key);
-    }
-    if (ctx->tag_key) {
-        flb_sds_destroy(ctx->tag_key);
-    }
-    if (ctx->json_date_key) {
-        flb_sds_destroy(ctx->json_date_key);
-    }
-    if (ctx->dd_source) {
-        flb_sds_destroy(ctx->dd_source);
-    }
-    if (ctx->dd_service) {
-        flb_sds_destroy(ctx->dd_service);
-    }
-    if (ctx->dd_tags) {
-        flb_sds_destroy(ctx->dd_tags);
-    }
-    if (ctx->dd_message_key) {
-        flb_sds_destroy(ctx->dd_message_key);
-    }
     if (ctx->upstream) {
         flb_upstream_destroy(ctx->upstream);
     }
