@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -28,6 +27,7 @@
 #include <fluent-bit/flb_macros.h>
 #include <fluent-bit/flb_config_map.h>
 #include <fluent-bit/flb_output_plugin.h>
+#include <fluent-bit/flb_log_event_decoder.h>
 
 #include <fluent-bit/flb_sds.h>
 #include <fluent-bit/flb_aws_credentials.h>
@@ -35,9 +35,9 @@
 #include <fluent-bit/flb_mem.h>
 #include <fluent-bit/flb_http_client.h>
 #include <fluent-bit/flb_utils.h>
+#include <fluent-bit/flb_base64.h>
 
 #include <monkey/mk_core.h>
-#include <mbedtls/base64.h>
 #include <msgpack.h>
 #include <string.h>
 #include <stdio.h>
@@ -226,6 +226,7 @@ static int process_event(struct flb_kinesis *ctx, struct flush *buf,
     struct tm *tmp;
     size_t len;
     size_t tmp_size;
+    char *out_buf;
 
     tmp_buf_ptr = buf->tmp_buf + buf->tmp_buf_offset;
     ret = flb_msgpack_to_json(tmp_buf_ptr,
@@ -277,34 +278,46 @@ static int process_event(struct flb_kinesis *ctx, struct flush *buf,
                          ctx->stream_name);
             return 2;
         }
-        /* guess space needed to write time_key */
-        len = 6 + strlen(ctx->time_key) + 6 * strlen(ctx->time_key_format);
+
+        /* format time output and return the length */
+        len = flb_aws_strftime_precision(&out_buf, ctx->time_key_format, tms);
+
         /* how much space do we have left */
         tmp_size = (buf->tmp_buf_size - buf->tmp_buf_offset) - written;
         if (len > tmp_size) {
-            /* not enough space- tell caller to retry */
+            /* not enough space - tell caller to retry */
+            flb_free(out_buf);
             return 1;
         }
-        time_key_ptr = tmp_buf_ptr + written - 1;
-        memcpy(time_key_ptr, ",", 1);
-        time_key_ptr++;
-        memcpy(time_key_ptr, "\"", 1);
-        time_key_ptr++;
-        memcpy(time_key_ptr, ctx->time_key, strlen(ctx->time_key));
-        time_key_ptr += strlen(ctx->time_key);
-        memcpy(time_key_ptr, "\":\"", 3);
-        time_key_ptr += 3;
-        tmp_size = buf->tmp_buf_size - buf->tmp_buf_offset;
-        tmp_size -= (time_key_ptr - tmp_buf_ptr);
-        len = strftime(time_key_ptr, tmp_size, ctx->time_key_format, &time_stamp);
-        if (len <= 0) {
-            /* ran out of space - should not happen because of check above */
-            return 1;
+
+        if (len == 0) {
+            /*
+             * when the length of out_buf is not enough for time_key_format,
+             * time_key will not be added to record.
+             */
+            flb_plg_error(ctx->ins, "Failed to add time_key %s to record, %s",
+                          ctx->time_key, ctx->stream_name);
+            flb_free(out_buf);
+        }
+        else {
+            time_key_ptr = tmp_buf_ptr + written - 1;
+            memcpy(time_key_ptr, ",", 1);
+            time_key_ptr++;
+            memcpy(time_key_ptr, "\"", 1);
+            time_key_ptr++;
+            memcpy(time_key_ptr, ctx->time_key, strlen(ctx->time_key));
+            time_key_ptr += strlen(ctx->time_key);
+            memcpy(time_key_ptr, "\":\"", 3);
+            time_key_ptr += 3;
+
+            /* merge out_buf to time_key_ptr */
+            memcpy(time_key_ptr, out_buf, len);
+            flb_free(out_buf);
+            time_key_ptr += len;
+            memcpy(time_key_ptr, "\"}", 2);
+            time_key_ptr += 2;
+            written = (time_key_ptr - tmp_buf_ptr);
         }
-        time_key_ptr += len;
-        memcpy(time_key_ptr, "\"}", 2);
-        time_key_ptr += 2;
-        written = (time_key_ptr - tmp_buf_ptr);
     }
 
     /* is (written + 1) because we still have to append newline */
@@ -342,7 +355,7 @@ static int process_event(struct flb_kinesis *ctx, struct flush *buf,
     }
 
     tmp_buf_ptr = buf->tmp_buf + buf->tmp_buf_offset;
-    ret = mbedtls_base64_encode((unsigned char *) buf->event_buf, size, &b64_len,
+    ret = flb_base64_encode((unsigned char *) buf->event_buf, size, &b64_len,
                                 (unsigned char *) tmp_buf_ptr, written);
     if (ret != 0) {
         flb_errno();
@@ -438,7 +451,7 @@ static int send_log_events(struct flb_kinesis *ctx, struct flush *buf) {
         flb_plg_error(ctx->ins, "Could not complete PutRecords payload");
         return -1;
     }
-    flb_plg_debug(ctx->ins, "Sending %d records", i);
+    flb_plg_debug(ctx->ins, "kinesis:PutRecords: events=%d, payload=%d bytes", i, offset);
     ret = put_records(ctx, buf, (size_t) offset, i);
     if (ret < 0) {
         flb_plg_error(ctx->ins, "Failed to send log records");
@@ -534,13 +547,9 @@ send:
 int process_and_send_to_kinesis(struct flb_kinesis *ctx, struct flush *buf,
                                const char *data, size_t bytes)
 {
-    size_t off = 0;
     int i = 0;
     size_t map_size;
-    msgpack_unpacked result;
-    msgpack_object  *obj;
     msgpack_object  map;
-    msgpack_object root;
     msgpack_object_kv *kv;
     msgpack_object  key;
     msgpack_object  val;
@@ -550,25 +559,22 @@ int process_and_send_to_kinesis(struct flb_kinesis *ctx, struct flush *buf,
     int ret;
     int check = FLB_FALSE;
     int found = FLB_FALSE;
-    struct flb_time tms;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
 
-    /* unpack msgpack */
-    msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
-        /*
-         * Each record is a msgpack array [timestamp, map] of the
-         * timestamp and record map.
-         */
-        root = result.data;
-        if (root.via.array.size != 2) {
-            continue;
-        }
+    ret = flb_log_event_decoder_init(&log_decoder, (char *) data, bytes);
 
-        /* unpack the array of [timestamp, map] */
-        flb_time_pop_from_msgpack(&tms, &result, &obj);
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
 
-        /* Get the record/map */
-        map = root.via.array.ptr[1];
+        return -1;
+    }
+
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
+        map = *log_event.body;
         map_size = map.via.map.size;
 
         if (ctx->log_key) {
@@ -596,7 +602,7 @@ int process_and_send_to_kinesis(struct flb_kinesis *ctx, struct flush *buf,
                     if (strncmp(ctx->log_key, key_str, key_str_size) == 0) {
                         found = FLB_TRUE;
                         val = (kv+j)->val;
-                        ret = add_event(ctx, buf, &val, &tms);
+                        ret = add_event(ctx, buf, &val, &log_event.timestamp);
                         if (ret < 0 ) {
                             goto error;
                         }
@@ -614,13 +620,14 @@ int process_and_send_to_kinesis(struct flb_kinesis *ctx, struct flush *buf,
             continue;
         }
 
-        ret = add_event(ctx, buf, &map, &tms);
+        ret = add_event(ctx, buf, &map, &log_event.timestamp);
         if (ret < 0 ) {
             goto error;
         }
         i++;
     }
-    msgpack_unpacked_destroy(&result);
+
+    flb_log_event_decoder_destroy(&log_decoder);
 
     /* send any remaining events */
     ret = send_log_events(ctx, buf);
@@ -634,7 +641,8 @@ int process_and_send_to_kinesis(struct flb_kinesis *ctx, struct flush *buf,
     return i;
 
 error:
-    msgpack_unpacked_destroy(&result);
+    flb_log_event_decoder_destroy(&log_decoder);
+
     return -1;
 }
 
@@ -668,7 +676,7 @@ static int process_api_response(struct flb_kinesis *ctx,
 
     /* Convert JSON payload to msgpack */
     ret = flb_pack_json(c->resp.payload, c->resp.payload_size,
-                        &out_buf, &out_size, &root_type);
+                        &out_buf, &out_size, &root_type, NULL);
     if (ret == -1) {
         flb_plg_error(ctx->ins, "could not pack/validate JSON API response\n%s",
                       c->resp.payload);
@@ -761,7 +769,7 @@ static int process_api_response(struct flb_kinesis *ctx,
                             (strncmp(response_val.via.str.ptr,
                                     ERR_CODE_EXCEEDED_THROUGHPUT, 38) == 0)) {
                                         throughput_exceeded = FLB_TRUE;
-                                        flb_plg_error(ctx->ins, "Thoughput limits may have been exceeded, %s",
+                                        flb_plg_error(ctx->ins, "Throughput limits may have been exceeded, %s",
                                                       ctx->stream_name);
                         }
                         flb_plg_debug(ctx->ins, "Record %i failed with err_code=%.*s",
