@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -21,6 +20,7 @@
 #include <fluent-bit/flb_output_plugin.h>
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_time.h>
+#include <fluent-bit/flb_log_event_decoder.h>
 #include <msgpack.h>
 
 #include <stdio.h>
@@ -38,7 +38,7 @@ static int cb_plot_init(struct flb_output_instance *ins,
                         struct flb_config *config,
                         void *data)
 {
-    const char *tmp;
+    int ret;
     (void) config;
     (void) data;
     struct flb_plot *ctx;
@@ -50,16 +50,10 @@ static int cb_plot_init(struct flb_output_instance *ins,
     }
     ctx->ins = ins;
 
-    /* Optional 'key' field to obtain the datapoint value */
-    tmp = flb_output_get_property("key", ins);
-    if (tmp) {
-        ctx->key = flb_sds_create(tmp);
-    }
-
-    /* Optional output file name/path */
-    tmp = flb_output_get_property("file", ins);
-    if (tmp) {
-        ctx->out_file = tmp;
+    ret = flb_output_config_map_set(ins, (void *)ctx);
+    if (ret == -1) {
+    	flb_free(ctx);
+    	return -1;
     }
 
     /* Set the context */
@@ -68,8 +62,8 @@ static int cb_plot_init(struct flb_output_instance *ins,
     return 0;
 }
 
-static void cb_plot_flush(const void *data, size_t bytes,
-                          const char *tag, int tag_len,
+static void cb_plot_flush(struct flb_event_chunk *event_chunk,
+                          struct flb_output_flush *out_flush,
                           struct flb_input_instance *i_ins,
                           void *out_context,
                           struct flb_config *config)
@@ -77,20 +71,21 @@ static void cb_plot_flush(const void *data, size_t bytes,
     int i;
     int written;
     int fd;
-    struct flb_time atime;
-    msgpack_unpacked result;
-    size_t off = 0;
     const char *out_file;
     msgpack_object *map;
     msgpack_object *key = NULL;
     msgpack_object *val = NULL;
     struct flb_plot *ctx = out_context;
+    struct flb_log_event_decoder log_decoder;
+    struct flb_log_event log_event;
+    int ret;
+
     (void) i_ins;
     (void) config;
 
     /* Set the right output */
     if (!ctx->out_file) {
-        out_file = tag;
+        out_file = event_chunk->tag;
     }
     else {
         out_file = ctx->out_file;
@@ -105,13 +100,27 @@ static void cb_plot_flush(const void *data, size_t bytes,
         fd = STDOUT_FILENO;
     }
 
+    ret = flb_log_event_decoder_init(&log_decoder,
+                                     (char *) event_chunk->data,
+                                     event_chunk->size);
+
+    if (ret != FLB_EVENT_DECODER_SUCCESS) {
+        flb_plg_error(ctx->ins,
+                      "Log event decoder initialization error : %d", ret);
+
+        close(fd);
+
+        FLB_OUTPUT_RETURN(FLB_ERROR);
+    }
+
     /*
      * Upon flush, for each array, lookup the time and the first field
      * of the map to use as a data point.
      */
-    msgpack_unpacked_init(&result);
-    while (msgpack_unpack_next(&result, data, bytes, &off) == MSGPACK_UNPACK_SUCCESS) {
-        flb_time_pop_from_msgpack(&atime, &result, &map);
+    while ((ret = flb_log_event_decoder_next(
+                    &log_decoder,
+                    &log_event)) == FLB_EVENT_DECODER_SUCCESS) {
+        map = log_event.body;
 
         /*
          * Lookup key, we need to iterate the whole map as sometimes the
@@ -146,6 +155,9 @@ static void cb_plot_flush(const void *data, size_t bytes,
                     if (fd != STDOUT_FILENO) {
                         close(fd);
                     }
+
+                    flb_log_event_decoder_destroy(&log_decoder);
+
                     FLB_OUTPUT_RETURN(FLB_ERROR);
                 }
             }
@@ -156,24 +168,27 @@ static void cb_plot_flush(const void *data, size_t bytes,
 
         if (!val) {
             flb_plg_error(ctx->ins, "unmatched key '%s'", ctx->key);
+
             if (fd != STDOUT_FILENO) {
                 close(fd);
             }
-            msgpack_unpacked_destroy(&result);
+
+            flb_log_event_decoder_destroy(&log_decoder);
+
             FLB_OUTPUT_RETURN(FLB_ERROR);
         }
 
         if (val->type == MSGPACK_OBJECT_POSITIVE_INTEGER) {
             written = dprintf(fd, "%f %" PRIu64 "\n",
-                              flb_time_to_double(&atime), val->via.u64);
+                              flb_time_to_double(&log_event.timestamp), val->via.u64);
         }
         else if (val->type == MSGPACK_OBJECT_NEGATIVE_INTEGER) {
             written = dprintf(fd, "%f %" PRId64 "\n",
-                              flb_time_to_double(&atime), val->via.i64);
+                              flb_time_to_double(&log_event.timestamp), val->via.i64);
         }
         else if (val->type == MSGPACK_OBJECT_FLOAT) {
             written = dprintf(fd, "%f %lf\n",
-                              flb_time_to_double(&atime), val->via.f64);
+                              flb_time_to_double(&log_event.timestamp), val->via.f64);
         }
         else {
             flb_plg_error(ctx->ins, "value must be integer, negative integer "
@@ -183,7 +198,8 @@ static void cb_plot_flush(const void *data, size_t bytes,
         flb_plg_debug(ctx->ins, "%i bytes written to file '%s'",
                       written, out_file);
     }
-    msgpack_unpacked_destroy(&result);
+
+    flb_log_event_decoder_destroy(&log_decoder);
 
     if (fd != STDOUT_FILENO) {
         close(fd);
@@ -196,17 +212,31 @@ static int cb_plot_exit(void *data, struct flb_config *config)
 {
     struct flb_plot *ctx = data;
 
-    flb_sds_destroy(ctx->key);
     flb_free(ctx);
-
     return 0;
 }
 
+static struct flb_config_map config_map[] = {
+    {
+     FLB_CONFIG_MAP_STR, "key", (char *)NULL,
+     0, FLB_TRUE, offsetof(struct flb_plot, key),
+     "set a number of times to generate event."
+    },
+    {
+     FLB_CONFIG_MAP_STR, "file", (char *)NULL,
+     0, FLB_TRUE, offsetof(struct flb_plot, out_file),
+     "set a number of times to generate event."
+    },
+    /* EOF */
+    {0}
+};
+
 struct flb_output_plugin out_plot_plugin = {
     .name         = "plot",
     .description  = "Generate data file for GNU Plot",
     .cb_init      = cb_plot_init,
     .cb_flush     = cb_plot_flush,
     .cb_exit      = cb_plot_exit,
+    .config_map   = config_map,
     .flags        = 0,
 };
