@@ -2,7 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -25,6 +25,7 @@
 #include <fluent-bit/flb_aws_credentials.h>
 #include <fluent-bit/flb_output_plugin.h>
 #include <fluent-bit/flb_jsmn.h>
+#include <fluent-bit/flb_env.h>
 
 #include <stdlib.h>
 #include <sys/types.h>
@@ -40,8 +41,28 @@
 #define S3_KEY_SIZE 1024
 #define RANDOM_STRING "$UUID"
 #define INDEX_STRING "$INDEX"
+#define AWS_USER_AGENT_NONE "none"
+#define AWS_USER_AGENT_ECS "ecs"
+#define AWS_USER_AGENT_K8S "k8s"
+#define AWS_ECS_METADATA_URI "ECS_CONTAINER_METADATA_URI_V4"
 #define FLB_MAX_AWS_RESP_BUFFER_SIZE 0 /* 0 means unlimited capacity as per requirement */
 
+#ifdef FLB_SYSTEM_WINDOWS
+#define FLB_AWS_BASE_USER_AGENT        "aws-fluent-bit-plugin-windows"
+#define FLB_AWS_BASE_USER_AGENT_FORMAT "aws-fluent-bit-plugin-windows-%s"
+#define FLB_AWS_BASE_USER_AGENT_LEN    29
+#else
+#define FLB_AWS_BASE_USER_AGENT        "aws-fluent-bit-plugin"
+#define FLB_AWS_BASE_USER_AGENT_FORMAT "aws-fluent-bit-plugin-%s"
+#define FLB_AWS_BASE_USER_AGENT_LEN    21
+#endif
+
+#define FLB_AWS_MILLISECOND_FORMATTER_LENGTH 3
+#define FLB_AWS_NANOSECOND_FORMATTER_LENGTH 9
+#define FLB_AWS_MILLISECOND_FORMATTER "%3N"
+#define FLB_AWS_NANOSECOND_FORMATTER_N "%9N"
+#define FLB_AWS_NANOSECOND_FORMATTER_L "%L"
+
 struct flb_http_client *request_do(struct flb_aws_client *aws_client,
                                    int method, const char *uri,
                                    const char *body, size_t body_len,
@@ -73,7 +94,7 @@ char *flb_aws_endpoint(char* service, char* region)
     len += strlen(region);
     len++; /* null byte */
 
-    endpoint = flb_malloc(len);
+    endpoint = flb_calloc(len, sizeof(char));
     if (!endpoint) {
         flb_errno();
         return NULL;
@@ -115,7 +136,7 @@ int flb_read_file(const char *path, char **out_buf, size_t *out_size)
         return -1;
     }
 
-    buf = flb_malloc(st.st_size + sizeof(char));
+    buf = flb_calloc(st.st_size + 1, sizeof(char));
     if (!buf) {
         flb_errno();
         close(fd);
@@ -178,7 +199,7 @@ struct flb_http_client *flb_aws_client_request(struct flb_aws_client *aws_client
         if (aws_client->has_auth && time(NULL) > aws_client->refresh_limit) {
             if (flb_aws_is_auth_error(c->resp.payload, c->resp.payload_size)
                 == FLB_TRUE) {
-                flb_error("[aws_client] auth error, refreshing creds");
+                flb_info("[aws_client] auth error, refreshing creds");
                 aws_client->refresh_limit = time(NULL)
                                             + FLB_AWS_CREDENTIAL_REFRESH_LIMIT;
                 aws_client->provider->provider_vtable->
@@ -224,6 +245,9 @@ void flb_aws_client_destroy(struct flb_aws_client *aws_client)
         if (aws_client->upstream) {
             flb_upstream_destroy(aws_client->upstream);
         }
+        if (aws_client->extra_user_agent) {
+            flb_sds_destroy(aws_client->extra_user_agent);
+        }
         flb_free(aws_client);
     }
 }
@@ -280,7 +304,7 @@ struct flb_http_client *request_do(struct flb_aws_client *aws_client,
 {
     size_t b_sent;
     int ret;
-    struct flb_upstream_conn *u_conn = NULL;
+    struct flb_connection *u_conn = NULL;
     flb_sds_t signature = NULL;
     int i;
     int normalize_uri;
@@ -288,6 +312,9 @@ struct flb_http_client *request_do(struct flb_aws_client *aws_client,
     struct flb_http_client *c = NULL;
     flb_sds_t tmp;
     flb_sds_t user_agent_prefix;
+    flb_sds_t user_agent = NULL;
+    char *buf;
+    struct flb_env *env;
 
     u_conn = flb_upstream_conn_get(aws_client->upstream);
     if (!u_conn) {
@@ -322,19 +349,55 @@ struct flb_http_client *request_do(struct flb_aws_client *aws_client,
         flb_warn("[aws_http_client] failed to increase max response buffer size");
     }
 
-    /* Add AWS Fluent Bit user agent */
+    /* Set AWS Fluent Bit user agent */
+    env = aws_client->upstream->base.config->env;
+    buf = (char *) flb_env_get(env, "FLB_AWS_USER_AGENT");
+    if (buf == NULL) {
+        if (getenv(AWS_ECS_METADATA_URI) != NULL) {
+            user_agent = AWS_USER_AGENT_ECS;
+        }
+        else {
+            buf = (char *) flb_env_get(env, AWS_USER_AGENT_K8S);
+            if (buf && strcasecmp(buf, "enabled") == 0) {
+                user_agent = AWS_USER_AGENT_K8S;
+            }
+        }
+
+        if (user_agent == NULL) {
+            user_agent = AWS_USER_AGENT_NONE;
+        }
+
+        flb_env_set(env, "FLB_AWS_USER_AGENT", user_agent);
+    }
     if (aws_client->extra_user_agent == NULL) {
+        buf = (char *) flb_env_get(env, "FLB_AWS_USER_AGENT");
+        tmp = flb_sds_create(buf);
+        if (!tmp) {
+            flb_errno();
+            goto error;
+        }
+        aws_client->extra_user_agent = tmp;
+        tmp = NULL;
+    }
+
+    /* Add AWS Fluent Bit user agent header */
+    if (strcasecmp(aws_client->extra_user_agent, AWS_USER_AGENT_NONE) == 0) {
         ret = flb_http_add_header(c, "User-Agent", 10,
-                                  "aws-fluent-bit-plugin", 21);
+                                  FLB_AWS_BASE_USER_AGENT, FLB_AWS_BASE_USER_AGENT_LEN);
     }
     else {
         user_agent_prefix = flb_sds_create_size(64);
-        tmp = flb_sds_printf(&user_agent_prefix, "aws-fluent-bit-plugin-%s",
+        if (!user_agent_prefix) {
+            flb_errno();
+            flb_error("[aws_client] failed to create user agent");
+            goto error;
+        }
+        tmp = flb_sds_printf(&user_agent_prefix, FLB_AWS_BASE_USER_AGENT_FORMAT,
                              aws_client->extra_user_agent);
         if (!tmp) {
             flb_errno();
             flb_sds_destroy(user_agent_prefix);
-            flb_error("[aws_client] failed to fetch user agent");
+            flb_error("[aws_client] failed to create user agent");
             goto error;
         }
         user_agent_prefix = tmp;
@@ -396,7 +459,7 @@ struct flb_http_client *request_do(struct flb_aws_client *aws_client,
         }
         signature = flb_signv4_do(c, normalize_uri, FLB_TRUE, time(NULL),
                                   aws_client->region, aws_client->service,
-                                  aws_client->s3_mode,
+                                  aws_client->s3_mode, NULL,
                                   aws_client->provider);
         if (!signature) {
             if (aws_client->debug_only == FLB_TRUE) {
@@ -445,13 +508,13 @@ void flb_aws_print_xml_error(char *response, size_t response_len,
     flb_sds_t error;
     flb_sds_t message;
 
-    error = flb_xml_get_val(response, response_len, "<Code>");
+    error = flb_aws_xml_get_val(response, response_len, "<Code>", "</Code>");
     if (!error) {
         flb_plg_error(ins, "%s: Could not parse response", api);
         return;
     }
 
-    message = flb_xml_get_val(response, response_len, "<Message>");
+    message = flb_aws_xml_get_val(response, response_len, "<Message>", "</Message>");
     if (!message) {
         /* just print the error */
         flb_plg_error(ins, "%s API responded with error='%s'", api, error);
@@ -468,14 +531,15 @@ void flb_aws_print_xml_error(char *response, size_t response_len,
 /* Parses AWS XML API Error responses and returns the value of the <code> tag */
 flb_sds_t flb_aws_xml_error(char *response, size_t response_len)
 {
-    return flb_xml_get_val(response, response_len, "<code>");
+    return flb_aws_xml_get_val(response, response_len, "<Code>", "</Code>");
 }
 
 /*
  * Parses an XML document and returns the value of the given tag
  * Param `tag` should include angle brackets; ex "<code>"
+ * And param `end` should include end brackets: "</code>"
  */
-flb_sds_t flb_xml_get_val(char *response, size_t response_len, char *tag)
+flb_sds_t flb_aws_xml_get_val(char *response, size_t response_len, char *tag, char *tag_end)
 {
     flb_sds_t val = NULL;
     char *node = NULL;
@@ -494,7 +558,7 @@ flb_sds_t flb_xml_get_val(char *response, size_t response_len, char *tag)
     /* advance to end of tag */
     node += strlen(tag);
 
-    end = strchr(node, '<');
+    end = strstr(node, tag_end);
     if (!end) {
         flb_error("[aws] Could not find end of '%s' node in xml", tag);
         return NULL;
@@ -637,7 +701,7 @@ static char* replace_uri_tokens(const char* original_string, const char* current
     i = 0;
     while (*original_string) {
         if (strstr(original_string, current_word) == original_string) {
-            strcpy(&result[i], new_word);
+            strncpy(&result[i], new_word, new_word_len);
             i += new_word_len;
             original_string += old_word_len;
         }
@@ -649,6 +713,23 @@ static char* replace_uri_tokens(const char* original_string, const char* current
     return result;
 }
 
+/*
+ * Linux has strtok_r as the concurrent safe version
+ * Windows has strtok_s
+ */
+char* strtok_concurrent(
+    char* str,
+    char* delimiters,
+    char** context
+)
+{
+#ifdef FLB_SYSTEM_WINDOWS
+    return strtok_s(str, delimiters, context);
+#else
+    return strtok_r(str, delimiters, context);
+#endif
+}
+
 /* Constructs S3 object key as per the format. */
 flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
                          char *tag_delimiter, uint64_t seq_index)
@@ -660,13 +741,15 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
     char *key;
     char *random_alphanumeric;
     char *seq_index_str;
+    /* concurrent safe strtok_r requires a tracking ptr */
+    char *strtok_saveptr;
     int len;
     flb_sds_t tmp = NULL;
     flb_sds_t buf = NULL;
     flb_sds_t s3_key = NULL;
     flb_sds_t tmp_key = NULL;
     flb_sds_t tmp_tag = NULL;
-    struct tm *gmt = NULL;
+    struct tm gmt = {0};
 
     if (strlen(format) > S3_KEY_SIZE){
         flb_warn("[s3_key] Object key length is longer than the 1024 character limit.");
@@ -705,7 +788,7 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
     tmp = NULL;
 
     /* Split the string on the delimiters */
-    tag_token = strtok(tmp_tag, tag_delimiter);
+    tag_token = strtok_concurrent(tmp_tag, tag_delimiter, &strtok_saveptr);
 
     /* Find all occurences of $TAG[*] and
      * replaces it with the right token from tag.
@@ -730,13 +813,17 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
             flb_warn("[s3_key] Object key length is longer than the 1024 character limit.");
         }
 
+        if (buf != tmp) {
+            flb_sds_destroy(buf);
+        }
         flb_sds_destroy(tmp);
         tmp = NULL;
+        buf = NULL;
         flb_sds_destroy(s3_key);
         s3_key = tmp_key;
         tmp_key = NULL;
 
-        tag_token = strtok(NULL, tag_delimiter);
+        tag_token = strtok_concurrent(NULL, tag_delimiter, &strtok_saveptr);
         i++;
     }
 
@@ -769,7 +856,7 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
     /* Find all occurences of $INDEX and replace with the appropriate index. */
     if (strstr((char *) format, INDEX_STRING)) {
         seq_index_len = snprintf(NULL, 0, "%"PRIu64, seq_index);
-        seq_index_str = flb_malloc(seq_index_len + 1);
+        seq_index_str = flb_calloc(seq_index_len + 1, sizeof(char));
         if (seq_index_str == NULL) {
             goto error;
         }
@@ -777,7 +864,10 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
         sprintf(seq_index_str, "%"PRIu64, seq_index);
         seq_index_str[seq_index_len] = '\0';
         tmp_key = replace_uri_tokens(s3_key, INDEX_STRING, seq_index_str);
-
+        if (tmp_key == NULL) {
+            flb_free(seq_index_str);
+            goto error;
+        }
         if (strlen(tmp_key) > S3_KEY_SIZE) {
             flb_warn("[s3_key] Object key length is longer than the 1024 character limit.");
         }
@@ -800,28 +890,31 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
         flb_free(random_alphanumeric);
         goto error;
     }
-    
+
     if(strlen(tmp_key) > S3_KEY_SIZE){
         flb_warn("[s3_key] Object key length is longer than the 1024 character limit.");
     }
-    
+
     flb_sds_destroy(s3_key);
     s3_key = tmp_key;
     tmp_key = NULL;
     flb_free(random_alphanumeric);
 
-    gmt = gmtime(&time);
+    if (!gmtime_r(&time, &gmt)) {
+        flb_error("[s3_key] Failed to create timestamp.");
+        goto error;
+    }
 
     flb_sds_destroy(tmp);
     tmp = NULL;
 
-    /* A string no longer than S3_KEY_SIZE is created to store the formatted timestamp. */
-    key = flb_calloc(1, S3_KEY_SIZE * sizeof(char));
+    /* A string no longer than S3_KEY_SIZE + 1 is created to store the formatted timestamp. */
+    key = flb_calloc(1, (S3_KEY_SIZE + 1) * sizeof(char));
     if (!key) {
         goto error;
     }
 
-    ret = strftime(key, S3_KEY_SIZE, s3_key, gmt);
+    ret = strftime(key, S3_KEY_SIZE, s3_key, &gmt);
     if(ret == 0){
         flb_warn("[s3_key] Object key length is longer than the 1024 character limit.");
     }
@@ -850,7 +943,7 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
         if (s3_key){
             flb_sds_destroy(s3_key);
         }
-        if (buf){
+        if (buf && buf != tmp){
             flb_sds_destroy(buf);
         }
         if (tmp){
@@ -861,3 +954,94 @@ flb_sds_t flb_get_s3_key(const char *format, time_t time, const char *tag,
         }
         return NULL;
 }
+
+/*
+ * This function is an extension to strftime which can support milliseconds with %3N,
+ * support nanoseconds with %9N or %L. The return value is the length of formatted
+ * time string.
+ */
+size_t flb_aws_strftime_precision(char **out_buf, const char *time_format,
+                                  struct flb_time *tms)
+{
+    char millisecond_str[FLB_AWS_MILLISECOND_FORMATTER_LENGTH+1];
+    char nanosecond_str[FLB_AWS_NANOSECOND_FORMATTER_LENGTH+1];
+    char *tmp_parsed_time_str;
+    char *buf;
+    size_t out_size;
+    size_t tmp_parsed_time_str_len;
+    size_t time_format_len;
+    struct tm timestamp;
+    struct tm *tmp;
+    int i;
+
+    /*
+     * Guess the max length needed for tmp_parsed_time_str and tmp_out_buf. The
+     * upper bound is 12*strlen(time_format) because the worst scenario will be only
+     * %c in time_format, and %c will be transfer to 24 chars long by function strftime().
+     */
+    time_format_len = strlen(time_format);
+    tmp_parsed_time_str_len = 12*time_format_len;
+
+    /*
+     * Use tmp_parsed_time_str to buffer when replace %3N with milliseconds, replace
+     * %9N and %L with nanoseconds in time_format.
+     */
+    tmp_parsed_time_str = (char *)flb_calloc(1, tmp_parsed_time_str_len*sizeof(char));
+    if (!tmp_parsed_time_str) {
+        flb_errno();
+        return 0;
+    }
+
+    buf = (char *)flb_calloc(1, tmp_parsed_time_str_len*sizeof(char));
+    if (!buf) {
+        flb_errno();
+        flb_free(tmp_parsed_time_str);
+        return 0;
+    }
+
+    /* Replace %3N to millisecond, %9N and %L to nanosecond in time_format. */
+    snprintf(millisecond_str, FLB_AWS_MILLISECOND_FORMATTER_LENGTH+1,
+             "%" PRIu64, (uint64_t) tms->tm.tv_nsec / 1000000);
+    snprintf(nanosecond_str, FLB_AWS_NANOSECOND_FORMATTER_LENGTH+1,
+             "%" PRIu64, (uint64_t) tms->tm.tv_nsec);
+    for (i = 0; i < time_format_len; i++) {
+        if (strncmp(time_format+i, FLB_AWS_MILLISECOND_FORMATTER, 3) == 0) {
+            strncat(tmp_parsed_time_str, millisecond_str,
+                    FLB_AWS_MILLISECOND_FORMATTER_LENGTH+1);
+            i += 2;
+        }
+        else if (strncmp(time_format+i, FLB_AWS_NANOSECOND_FORMATTER_N, 3) == 0) {
+            strncat(tmp_parsed_time_str, nanosecond_str,
+                    FLB_AWS_NANOSECOND_FORMATTER_LENGTH+1);
+            i += 2;
+        }
+        else if (strncmp(time_format+i, FLB_AWS_NANOSECOND_FORMATTER_L, 2) == 0) {
+            strncat(tmp_parsed_time_str, nanosecond_str,
+                    FLB_AWS_NANOSECOND_FORMATTER_LENGTH+1);
+            i += 1;
+        }
+        else {
+            strncat(tmp_parsed_time_str,time_format+i,1);
+        }
+    }
+
+    tmp = gmtime_r(&tms->tm.tv_sec, &timestamp);
+    if (!tmp) {
+        return 0;
+    }
+
+    out_size = strftime(buf, tmp_parsed_time_str_len,
+                        tmp_parsed_time_str, &timestamp);
+
+    /* Check whether tmp_parsed_time_str_len is enough for tmp_out_buff */
+    if (out_size == 0) {
+        flb_free(tmp_parsed_time_str);
+        flb_free(buf);
+        return 0;
+    }
+
+    *out_buf = buf;
+    flb_free(tmp_parsed_time_str);
+
+    return out_size;
+}
