@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -34,17 +33,18 @@
 #include <fluent-bit/flb_config.h>
 #include <fluent-bit/flb_macros.h>
 #include <fluent-bit/flb_utils.h>
+#include <fluent-bit/flb_plugin.h>
 #include <fluent-bit/flb_plugin_proxy.h>
 #include <fluent-bit/flb_http_client_debug.h>
 #include <fluent-bit/flb_output_thread.h>
 #include <fluent-bit/flb_mp.h>
 #include <fluent-bit/flb_pack.h>
 
-FLB_TLS_DEFINE(struct flb_out_coro_params, out_coro_params);
+FLB_TLS_DEFINE(struct flb_out_flush_params, out_flush_params);
 
 void flb_output_prepare()
 {
-    FLB_TLS_INIT(out_coro_params);
+    FLB_TLS_INIT(out_flush_params);
 }
 
 /* Validate the the output address protocol */
@@ -118,26 +118,26 @@ static void flb_output_free_properties(struct flb_output_instance *ins)
 #endif
 }
 
-void flb_output_coro_prepare_destroy(struct flb_output_coro *out_coro)
+void flb_output_flush_prepare_destroy(struct flb_output_flush *out_flush)
 {
-    struct flb_output_instance *ins = out_coro->o_ins;
+    struct flb_output_instance *ins = out_flush->o_ins;
     struct flb_out_thread_instance *th_ins;
 
     /* Move output coroutine context from active list to the destroy one */
     if (flb_output_is_threaded(ins) == FLB_TRUE) {
         th_ins = flb_output_thread_instance_get();
-        pthread_mutex_lock(&th_ins->coro_mutex);
-        mk_list_del(&out_coro->_head);
-        mk_list_add(&out_coro->_head, &th_ins->coros_destroy);
-        pthread_mutex_unlock(&th_ins->coro_mutex);
+        pthread_mutex_lock(&th_ins->flush_mutex);
+        mk_list_del(&out_flush->_head);
+        mk_list_add(&out_flush->_head, &th_ins->flush_list_destroy);
+        pthread_mutex_unlock(&th_ins->flush_mutex);
     }
     else {
-        mk_list_del(&out_coro->_head);
-        mk_list_add(&out_coro->_head, &ins->coros_destroy);
+        mk_list_del(&out_flush->_head);
+        mk_list_add(&out_flush->_head, &ins->flush_list_destroy);
     }
 }
 
-int flb_output_coro_id_get(struct flb_output_instance *ins)
+int flb_output_flush_id_get(struct flb_output_instance *ins)
 {
     int id;
     int max = (2 << 13) - 1; /* max for 14 bits */
@@ -145,21 +145,21 @@ int flb_output_coro_id_get(struct flb_output_instance *ins)
 
     if (flb_output_is_threaded(ins) == FLB_TRUE) {
         th_ins = flb_output_thread_instance_get();
-        id = th_ins->coro_id;
-        th_ins->coro_id++;
+        id = th_ins->flush_id;
+        th_ins->flush_id++;
 
         /* reset once it reach the maximum allowed */
-        if (th_ins->coro_id > max) {
-            th_ins->coro_id = 0;
+        if (th_ins->flush_id > max) {
+            th_ins->flush_id = 0;
         }
     }
     else {
-        id = ins->coro_id;
-        ins->coro_id++;
+        id = ins->flush_id;
+        ins->flush_id++;
 
         /* reset once it reach the maximum allowed */
-        if (ins->coro_id > max) {
-            ins->coro_id = 0;
+        if (ins->flush_id > max) {
+            ins->flush_id = 0;
         }
     }
 
@@ -168,10 +168,145 @@ int flb_output_coro_id_get(struct flb_output_instance *ins)
 
 void flb_output_coro_add(struct flb_output_instance *ins, struct flb_coro *coro)
 {
-    struct flb_output_coro *out_coro;
+    struct flb_output_flush *out_flush;
 
-    out_coro = (struct flb_output_coro *) FLB_CORO_DATA(coro);
-    mk_list_add(&out_coro->_head, &ins->coros);
+    out_flush = (struct flb_output_flush *) FLB_CORO_DATA(coro);
+    mk_list_add(&out_flush->_head, &ins->flush_list);
+}
+
+/*
+ * Queue a task to be flushed at a later time
+ * Deletes retry context if enqueue fails
+ */
+static int flb_output_task_queue_enqueue(struct flb_task_queue *queue,
+                                         struct flb_task_retry *retry,
+                                         struct flb_task *task,
+                                         struct flb_output_instance *out_ins,
+                                         struct flb_config *config)
+{
+    struct flb_task_enqueued *queued_task;
+
+    queued_task = flb_malloc(sizeof(struct flb_task_enqueued));
+    if (!queued_task) {
+        flb_errno();
+        if (retry) {
+            flb_task_retry_destroy(retry);
+        }
+        return -1;
+    }
+    queued_task->retry = retry;
+    queued_task->out_instance = out_ins;
+    queued_task->task = task;
+    queued_task->config = config;
+
+    mk_list_add(&queued_task->_head, &queue->pending);
+    return 0;
+}
+
+/*
+ * Pop task from pending queue and flush it
+ * Will delete retry context if flush fails
+ */
+static int flb_output_task_queue_flush_one(struct flb_task_queue *queue)
+{
+    struct flb_task_enqueued *queued_task;
+    int ret;
+    int is_empty;
+
+    is_empty = mk_list_is_empty(&queue->pending) == 0;
+    if (is_empty) {
+        flb_error("Attempting to flush task from an empty in_progress queue");
+        return -1;
+    }
+
+    queued_task = mk_list_entry_first(&queue->pending, struct flb_task_enqueued, _head);
+    mk_list_del(&queued_task->_head);
+    mk_list_add(&queued_task->_head, &queue->in_progress);
+
+    /*
+     * Remove temporary user now that task is out of singleplex queue.
+     * Flush will add back the user representing queued_task->out_instance if it succeeds.
+     */
+    flb_task_users_dec(queued_task->task, FLB_FALSE);
+    ret = flb_output_task_flush(queued_task->task,
+                                queued_task->out_instance,
+                                queued_task->config);
+
+    /* Destroy retry context if needed */
+    if (ret == -1) {
+        if (queued_task->retry) {
+            flb_task_retry_destroy(queued_task->retry);
+        }
+        /* Flush the next task */
+        flb_output_task_singleplex_flush_next(queue);
+        return -1;
+    }
+
+    return ret;
+}
+
+/*
+ * Will either run or queue running a single task
+ * Deletes retry context if enqueue fails
+ */
+int flb_output_task_singleplex_enqueue(struct flb_task_queue *queue,
+                                       struct flb_task_retry *retry,
+                                       struct flb_task *task,
+                                       struct flb_output_instance *out_ins,
+                                       struct flb_config *config)
+{
+    int ret;
+    int is_empty;
+
+    /*
+     * Add temporary user to preserve task while in singleplex queue.
+     * Temporary user will be removed when task is removed from queue.
+     *
+     * Note: if we fail to increment now, then the task may be prematurely
+     * deleted if the task's users go to 0 while we are waiting in the
+     * queue.
+     */
+    flb_task_users_inc(task);
+
+    /* Enqueue task */
+    ret = flb_output_task_queue_enqueue(queue, retry, task, out_ins, config);
+    if (ret == -1) {
+        return -1;
+    }
+
+    /* Launch task if nothing is running */
+    is_empty = mk_list_is_empty(&out_ins->singleplex_queue->in_progress) == 0;
+    if (is_empty) {
+        return flb_output_task_queue_flush_one(out_ins->singleplex_queue);
+    }
+    
+    return 0;
+}
+
+/*
+ * Clear in progress task and flush a single queued task if exists
+ * Deletes retry context on next flush if flush fails
+ */
+int flb_output_task_singleplex_flush_next(struct flb_task_queue *queue)
+{
+    int is_empty;
+    struct flb_task_enqueued *ended_task;
+
+    /* Remove in progress task */
+    is_empty = mk_list_is_empty(&queue->in_progress) == 0;
+    if (!is_empty) {
+        ended_task = mk_list_entry_first(&queue->in_progress,
+                                        struct flb_task_enqueued, _head);
+        mk_list_del(&ended_task->_head);
+        flb_free(ended_task);
+    }
+    
+    /* Flush if there is a pending task queued */
+    is_empty = mk_list_is_empty(&queue->pending) == 0;
+    if (!is_empty) {
+        return flb_output_task_queue_flush_one(queue);
+    }
+    return 0;
 }
 
 /*
@@ -183,7 +318,7 @@ int flb_output_task_flush(struct flb_task *task,
                           struct flb_config *config)
 {
     int ret;
-    struct flb_output_coro *out_coro;
+    struct flb_output_flush *out_flush;
 
     if (flb_output_is_threaded(out_ins) == FLB_TRUE) {
         flb_task_users_inc(task);
@@ -192,23 +327,38 @@ int flb_output_task_flush(struct flb_task *task,
         ret = flb_output_thread_pool_flush(task, out_ins, config);
         if (ret == -1) {
             flb_task_users_dec(task, FLB_FALSE);
+
+            /* If we are in synchronous mode, flush one waiting task */
+            if (out_ins->flags & FLB_OUTPUT_SYNCHRONOUS) {
+                flb_output_task_singleplex_flush_next(out_ins->singleplex_queue);
+            }
         }
     }
     else {
-        /* Direct co-routine handling */
-        out_coro = flb_output_coro_create(task,
-                                          task->i_ins,
-                                          out_ins,
-                                          config,
-                                          task->buf, task->size,
-                                          task->tag,
-                                          task->tag_len);
-        if (!out_coro) {
+        /* Queue co-routine handling */
+        out_flush = flb_output_flush_create(task,
+                                           task->i_ins,
+                                           out_ins,
+                                           config);
+        if (!out_flush) {
             return -1;
         }
 
         flb_task_users_inc(task);
-        flb_coro_resume(out_coro->coro);
+        ret = flb_pipe_w(config->ch_self_events[1], &out_flush,
+                        sizeof(struct flb_output_flush*));
+        if (ret == -1) {
+            flb_errno();
+            flb_output_flush_destroy(out_flush);
+            flb_task_users_dec(task, FLB_FALSE);
+
+            /* If we are in synchronous mode, flush one waiting task */
+            if (out_ins->flags & FLB_OUTPUT_SYNCHRONOUS) {
+                flb_output_task_singleplex_flush_next(out_ins->singleplex_queue);
+            }
+
+            return -1;
+        }
     }
 
     return 0;
@@ -284,7 +434,18 @@ int flb_output_instance_destroy(struct flb_output_instance *ins)
     /* release properties */
     flb_output_free_properties(ins);
 
+    /* free singleplex queue */
+    if (ins->flags & FLB_OUTPUT_SYNCHRONOUS) {
+        flb_task_queue_destroy(ins->singleplex_queue);
+    }
+
     mk_list_del(&ins->_head);
+
+    /* processor */
+    if (ins->processor) {
+        flb_processor_destroy(ins->processor);
+    }
+
     flb_free(ins);
 
     return 0;
@@ -310,17 +471,12 @@ void flb_output_exit(struct flb_config *config)
 
         /* Check a exit callback */
         if (p->cb_exit) {
-            if(!p->proxy) {
-                p->cb_exit(ins->context, config);
-            }
-            else {
-                p->cb_exit(p, ins->context);
-            }
+            p->cb_exit(ins->context, config);
         }
         flb_output_instance_destroy(ins);
     }
 
-    params = FLB_TLS_GET(out_coro_params);
+    params = FLB_TLS_GET(out_flush_params);
     if (params) {
         flb_free(params);
     }
@@ -370,7 +526,7 @@ int flb_output_flush_finished(struct flb_config *config, int out_id)
     struct mk_list *head;
     struct mk_list *list;
     struct flb_output_instance *ins;
-    struct flb_output_coro *out_coro;
+    struct flb_output_flush *out_flush;
     struct flb_out_thread_instance *th_ins;
 
     ins = flb_output_get_instance(config, out_id);
@@ -380,16 +536,16 @@ int flb_output_flush_finished(struct flb_config *config, int out_id)
 
     if (flb_output_is_threaded(ins) == FLB_TRUE) {
         th_ins = flb_output_thread_instance_get();
-        list = &th_ins->coros_destroy;
+        list = &th_ins->flush_list_destroy;
     }
     else {
-        list = &ins->coros_destroy;
+        list = &ins->flush_list_destroy;
     }
 
     /* Look for output coroutines that needs to be destroyed */
     mk_list_foreach_safe(head, tmp, list) {
-        out_coro = mk_list_entry(head, struct flb_output_coro, _head);
-        flb_output_coro_destroy(out_coro);
+        out_flush = mk_list_entry(head, struct flb_output_flush, _head);
+        flb_output_flush_destroy(out_flush);
     }
 
     return 0;
@@ -447,9 +603,10 @@ struct flb_output_instance *flb_output_new(struct flb_config *config,
     }
     instance->config = config;
     instance->log_level = -1;
+    instance->log_suppress_interval = -1;
     instance->test_mode = FLB_FALSE;
     instance->is_threaded = FLB_FALSE;
-
+    instance->tp_workers = plugin->workers;
 
     /* Retrieve an instance id for the output instance */
     instance->id = instance_id(config);
@@ -460,6 +617,9 @@ struct flb_output_instance *flb_output_new(struct flb_config *config,
     instance->p = plugin;
     instance->callback = flb_callback_create(instance->name);
     if (!instance->callback) {
+        if (instance->flags & FLB_OUTPUT_SYNCHRONOUS) {
+            flb_task_queue_destroy(instance->singleplex_queue);
+        }
         flb_free(instance);
         return NULL;
     }
@@ -473,6 +633,9 @@ struct flb_output_instance *flb_output_new(struct flb_config *config,
         ctx = flb_calloc(1, sizeof(struct flb_plugin_proxy_context));
         if (!ctx) {
             flb_errno();
+            if (instance->flags & FLB_OUTPUT_SYNCHRONOUS) {
+                flb_task_queue_destroy(instance->singleplex_queue);
+            }
             flb_free(instance);
             return NULL;
         }
@@ -526,22 +689,40 @@ struct flb_output_instance *flb_output_new(struct flb_config *config,
     if (plugin->flags & FLB_OUTPUT_NET) {
         ret = flb_net_host_set(plugin->name, &instance->host, output);
         if (ret != 0) {
+            if (instance->flags & FLB_OUTPUT_SYNCHRONOUS) {
+                flb_task_queue_destroy(instance->singleplex_queue);
+            }
             flb_free(instance);
             return NULL;
         }
     }
 
+    /* Create singleplex queue if SYNCHRONOUS mode is used */
+    instance->singleplex_queue = NULL;
+    if (instance->flags & FLB_OUTPUT_SYNCHRONOUS) {
+        instance->singleplex_queue = flb_task_queue_create();
+        if (!instance->singleplex_queue) {
+            flb_free(instance);
+            flb_errno();
+            return NULL;
+        }
+    }
+
     flb_kv_init(&instance->properties);
     flb_kv_init(&instance->net_properties);
     mk_list_init(&instance->upstreams);
-    mk_list_init(&instance->coros);
-    mk_list_init(&instance->coros_destroy);
+    mk_list_init(&instance->flush_list);
+    mk_list_init(&instance->flush_list_destroy);
 
     mk_list_add(&instance->_head, &config->outputs);
 
+    /* processor instance */
+    instance->processor = flb_processor_create(config, instance->name, instance, FLB_PLUGIN_OUTPUT);
+
     /* Tests */
     instance->test_formatter.callback = plugin->test_formatter.callback;
 
+
     return instance;
 }
 
@@ -579,7 +760,7 @@ int flb_output_set_property(struct flb_output_instance *ins,
 
     /* Check if the key is a known/shared property */
     if (prop_key_check("match", k, len) == 0) {
-        ins->match = tmp;
+        flb_utils_set_plugin_string_property("match", &ins->match, tmp);
     }
 #ifdef FLB_HAVE_REGEX
     else if (prop_key_check("match_regex", k, len) == 0 && tmp) {
@@ -588,7 +769,7 @@ int flb_output_set_property(struct flb_output_instance *ins,
     }
 #endif
     else if (prop_key_check("alias", k, len) == 0 && tmp) {
-        ins->alias = tmp;
+        flb_utils_set_plugin_string_property("alias", &ins->alias, tmp);
     }
     else if (prop_key_check("log_level", k, len) == 0 && tmp) {
         ret = flb_log_get_level_str(tmp);
@@ -598,8 +779,16 @@ int flb_output_set_property(struct flb_output_instance *ins,
         }
         ins->log_level = ret;
     }
+    else if (prop_key_check("log_suppress_interval", k, len) == 0 && tmp) {
+        ret = flb_utils_time_to_seconds(tmp);
+        flb_sds_destroy(tmp);
+        if (ret == -1) {
+            return -1;
+        }
+        ins->log_suppress_interval = ret;
+    }
     else if (prop_key_check("host", k, len) == 0) {
-        ins->host.name = tmp;
+        flb_utils_set_plugin_string_property("host", &ins->host.name, tmp);
     }
     else if (prop_key_check("port", k, len) == 0) {
         if (tmp) {
@@ -671,27 +860,16 @@ int flb_output_set_property(struct flb_output_instance *ins,
 #endif
 #ifdef FLB_HAVE_TLS
     else if (prop_key_check("tls", k, len) == 0 && tmp) {
-        if (strcasecmp(tmp, "true") == 0 || strcasecmp(tmp, "on") == 0) {
-            if ((ins->flags & FLB_IO_TLS) == 0) {
-                flb_error("[config] %s don't support TLS", ins->name);
-                flb_sds_destroy(tmp);
-                return -1;
-            }
-
-            ins->use_tls = FLB_TRUE;
-        }
-        else {
-            ins->use_tls = FLB_FALSE;
+        ins->use_tls = flb_utils_bool(tmp);
+        if (ins->use_tls == FLB_TRUE && ((ins->flags & FLB_IO_TLS) == 0)) {
+            flb_error("[config] %s does not support TLS", ins->name);
+            flb_sds_destroy(tmp);
+            return -1;
         }
         flb_sds_destroy(tmp);
     }
     else if (prop_key_check("tls.verify", k, len) == 0 && tmp) {
-        if (strcasecmp(tmp, "true") == 0 || strcasecmp(tmp, "on") == 0) {
-            ins->tls_verify = FLB_TRUE;
-        }
-        else {
-            ins->tls_verify = FLB_FALSE;
-        }
+        ins->tls_verify = flb_utils_bool(tmp);
         flb_sds_destroy(tmp);
     }
     else if (prop_key_check("tls.debug", k, len) == 0 && tmp) {
@@ -699,22 +877,22 @@ int flb_output_set_property(struct flb_output_instance *ins,
         flb_sds_destroy(tmp);
     }
     else if (prop_key_check("tls.vhost", k, len) == 0) {
-        ins->tls_vhost = tmp;
+        flb_utils_set_plugin_string_property("tls.vhost", &ins->tls_vhost, tmp);
     }
     else if (prop_key_check("tls.ca_path", k, len) == 0) {
-        ins->tls_ca_path = tmp;
+        flb_utils_set_plugin_string_property("tls.ca_path", &ins->tls_ca_path, tmp);
     }
     else if (prop_key_check("tls.ca_file", k, len) == 0) {
-        ins->tls_ca_file = tmp;
+        flb_utils_set_plugin_string_property("tls.ca_file", &ins->tls_ca_file, tmp);
     }
     else if (prop_key_check("tls.crt_file", k, len) == 0) {
-        ins->tls_crt_file = tmp;
+        flb_utils_set_plugin_string_property("tls.crt_file", &ins->tls_crt_file, tmp);
     }
     else if (prop_key_check("tls.key_file", k, len) == 0) {
-        ins->tls_key_file = tmp;
+        flb_utils_set_plugin_string_property("tls.key_file", &ins->tls_key_file, tmp);
     }
     else if (prop_key_check("tls.key_passwd", k, len) == 0) {
-        ins->tls_key_passwd = tmp;
+        flb_utils_set_plugin_string_property("tls.key_passwd", &ins->tls_key_passwd, tmp);
     }
 #endif
     else if (prop_key_check("storage.total_limit_size", k, len) == 0 && tmp) {
@@ -806,6 +984,81 @@ const char *flb_output_get_property(const char *key, struct flb_output_instance
     return flb_config_prop_get(key, &ins->properties);
 }
 
+#ifdef FLB_HAVE_METRICS
+void *flb_output_get_cmt_instance(struct flb_output_instance *ins)
+{
+    return (void *)ins->cmt;
+}
+#endif
+
+int flb_output_net_property_check(struct flb_output_instance *ins,
+                                  struct flb_config *config)
+{
+    int ret = 0;
+
+    /* Get Upstream net_setup configmap */
+    ins->net_config_map = flb_upstream_get_config_map(config);
+    if (!ins->net_config_map) {
+        flb_output_instance_destroy(ins);
+        return -1;
+    }
+
+    /*
+     * Validate 'net.*' properties: if the plugin use the Upstream interface,
+     * it might receive some networking settings.
+     */
+    if (mk_list_size(&ins->net_properties) > 0) {
+        ret = flb_config_map_properties_check(ins->p->name,
+                                              &ins->net_properties,
+                                              ins->net_config_map);
+        if (ret == -1) {
+            if (config->program_name) {
+                flb_helper("try the command: %s -o %s -h\n",
+                           config->program_name, ins->p->name);
+            }
+            return -1;
+        }
+    }
+
+    return 0;
+}
+
+int flb_output_plugin_property_check(struct flb_output_instance *ins,
+                                     struct flb_config *config)
+{
+    int ret = 0;
+    struct mk_list *config_map;
+    struct flb_output_plugin *p = ins->p;
+
+    if (p->config_map) {
+        /*
+         * Create a dynamic version of the configmap that will be used by the specific
+         * instance in question.
+         */
+        config_map = flb_config_map_create(config, p->config_map);
+        if (!config_map) {
+            flb_error("[output] error loading config map for '%s' plugin",
+                      p->name);
+            flb_output_instance_destroy(ins);
+            return -1;
+        }
+        ins->config_map = config_map;
+
+        /* Validate incoming properties against config map */
+        ret = flb_config_map_properties_check(ins->p->name,
+                                              &ins->properties, ins->config_map);
+        if (ret == -1) {
+            if (config->program_name) {
+                flb_helper("try the command: %s -o %s -h\n",
+                           config->program_name, ins->p->name);
+            }
+            return -1;
+        }
+    }
+
+    return 0;
+}
+
 /* Trigger the output plugins setup callbacks to prepare them. */
 int flb_output_init_all(struct flb_config *config)
 {
@@ -815,9 +1068,9 @@ int flb_output_init_all(struct flb_config *config)
 #endif
     struct mk_list *tmp;
     struct mk_list *head;
-    struct mk_list *config_map;
     struct flb_output_instance *ins;
     struct flb_output_plugin *p;
+    uint64_t ts;
 
     /* Retrieve the plugin reference */
     mk_list_foreach_safe(head, tmp, &config->outputs) {
@@ -854,6 +1107,9 @@ int flb_output_init_all(struct flb_config *config)
         /* Get name or alias for the instance */
         name = (char *) flb_output_name(ins);
 
+        /* get timestamp */
+        ts = cfl_time_now();
+
         /* CMetrics */
         ins->cmt = cmt_create();
         if (!ins->cmt) {
@@ -861,43 +1117,88 @@ int flb_output_init_all(struct flb_config *config)
             return -1;
         }
 
-        /* Register generic output plugin metrics */
+        /*
+         * Register generic output plugin metrics
+         */
+
+        /* fluentbit_output_proc_records_total */
         ins->cmt_proc_records = cmt_counter_create(ins->cmt, "fluentbit",
                                                    "output", "proc_records_total",
                                                    "Number of processed output records.",
                                                    1, (char *[]) {"name"});
+        cmt_counter_set(ins->cmt_proc_records, ts, 0, 1, (char *[]) {name});
 
+
+        /* fluentbit_output_proc_bytes_total */
         ins->cmt_proc_bytes = cmt_counter_create(ins->cmt, "fluentbit",
                                                  "output", "proc_bytes_total",
                                                  "Number of processed output bytes.",
                                                  1, (char *[]) {"name"});
+        cmt_counter_set(ins->cmt_proc_bytes, ts, 0, 1, (char *[]) {name});
 
+
+        /* fluentbit_output_errors_total */
         ins->cmt_errors = cmt_counter_create(ins->cmt, "fluentbit",
                                              "output", "errors_total",
                                              "Number of output errors.",
                                              1, (char *[]) {"name"});
+        cmt_counter_set(ins->cmt_errors, ts, 0, 1, (char *[]) {name});
 
+
+        /* fluentbit_output_retries_total */
         ins->cmt_retries = cmt_counter_create(ins->cmt, "fluentbit",
                                              "output", "retries_total",
                                              "Number of output retries.",
                                              1, (char *[]) {"name"});
+        cmt_counter_set(ins->cmt_retries, ts, 0, 1, (char *[]) {name});
 
+        /* fluentbit_output_retries_failed_total */
         ins->cmt_retries_failed = cmt_counter_create(ins->cmt, "fluentbit",
                                              "output", "retries_failed_total",
                                              "Number of abandoned batches because "
                                              "the maximum number of re-tries was "
                                              "reached.",
                                              1, (char *[]) {"name"});
+        cmt_counter_set(ins->cmt_retries_failed, ts, 0, 1, (char *[]) {name});
 
+
+        /* fluentbit_output_dropped_records_total */
         ins->cmt_dropped_records = cmt_counter_create(ins->cmt, "fluentbit",
                                              "output", "dropped_records_total",
                                              "Number of dropped records.",
                                              1, (char *[]) {"name"});
+        cmt_counter_set(ins->cmt_dropped_records, ts, 0, 1, (char *[]) {name});
 
+        /* fluentbit_output_retried_records_total */
         ins->cmt_retried_records = cmt_counter_create(ins->cmt, "fluentbit",
                                              "output", "retried_records_total",
                                              "Number of retried records.",
                                              1, (char *[]) {"name"});
+        cmt_counter_set(ins->cmt_retried_records, ts, 0, 1, (char *[]) {name});
+
+        /* output_upstream_total_connections */
+        ins->cmt_upstream_total_connections = cmt_gauge_create(ins->cmt,
+                                                               "fluentbit",
+                                                               "output",
+                                                               "upstream_total_connections",
+                                                               "Total Connection count.",
+                                                               1, (char *[]) {"name"});
+        cmt_gauge_set(ins->cmt_upstream_total_connections,
+                      ts,
+                      0,
+                      1, (char *[]) {name});
+
+        /* output_upstream_total_connections */
+        ins->cmt_upstream_busy_connections = cmt_gauge_create(ins->cmt,
+                                                              "fluentbit",
+                                                              "output",
+                                                              "upstream_busy_connections",
+                                                              "Busy Connection count.",
+                                                              1, (char *[]) {"name"});
+        cmt_gauge_set(ins->cmt_upstream_busy_connections,
+                      ts,
+                      0,
+                      1, (char *[]) {name});
 
         /* old API */
         ins->metrics = flb_metrics_create(name);
@@ -922,7 +1223,7 @@ int flb_output_init_all(struct flb_config *config)
 #ifdef FLB_HAVE_PROXY_GO
         /* Proxy plugins have their own initialization */
         if (p->type == FLB_OUTPUT_PLUGIN_PROXY) {
-            ret = flb_plugin_proxy_init(p->proxy, ins, config);
+            ret = flb_plugin_proxy_output_init(p->proxy, ins, config);
             if (ret == -1) {
                 flb_output_instance_destroy(ins);
                 return -1;
@@ -942,7 +1243,8 @@ int flb_output_init_all(struct flb_config *config)
 
 #ifdef FLB_HAVE_TLS
         if (ins->use_tls == FLB_TRUE) {
-            ins->tls = flb_tls_create(ins->tls_verify,
+            ins->tls = flb_tls_create(FLB_TLS_CLIENT_MODE,
+                                      ins->tls_verify,
                                       ins->tls_debug,
                                       ins->tls_vhost,
                                       ins->tls_ca_path,
@@ -962,39 +1264,7 @@ int flb_output_init_all(struct flb_config *config)
          * Before to call the initialization callback, make sure that the received
          * configuration parameters are valid if the plugin is registering a config map.
          */
-        if (p->config_map) {
-            /*
-             * Create a dynamic version of the configmap that will be used by the specific
-             * instance in question.
-             */
-            config_map = flb_config_map_create(config, p->config_map);
-            if (!config_map) {
-                flb_error("[output] error loading config map for '%s' plugin",
-                          p->name);
-                flb_output_instance_destroy(ins);
-                return -1;
-            }
-            ins->config_map = config_map;
-
-            /* Validate incoming properties against config map */
-            ret = flb_config_map_properties_check(ins->p->name,
-                                                  &ins->properties, ins->config_map);
-            if (ret == -1) {
-                if (config->program_name) {
-                    flb_helper("try the command: %s -o %s -h\n",
-                               config->program_name, ins->p->name);
-                }
-                flb_output_instance_destroy(ins);
-                return -1;
-            }
-        }
-
-        /* Init network defaults */
-        flb_net_setup_init(&ins->net_setup);
-
-        /* Get Upstream net_setup configmap */
-        ins->net_config_map = flb_upstream_get_config_map(config);
-        if (!ins->net_config_map) {
+        if (flb_output_plugin_property_check(ins, config) == -1) {
             flb_output_instance_destroy(ins);
             return -1;
         }
@@ -1004,6 +1274,10 @@ int flb_output_init_all(struct flb_config *config)
 
         /* TLS config map (just for 'help' formatting purposes) */
         ins->tls_config_map = flb_tls_get_config_map(config);
+        if (!ins->tls_config_map) {
+            flb_output_instance_destroy(ins);
+            return -1;
+        }
 
         /* Override first configmap value based on it plugin flag */
         m = mk_list_entry_first(ins->tls_config_map, struct flb_config_map, _head);
@@ -1014,22 +1288,13 @@ int flb_output_init_all(struct flb_config *config)
             m->value.val.boolean = FLB_FALSE;
         }
 #endif
-        /*
-         * Validate 'net.*' properties: if the plugin use the Upstream interface,
-         * it might receive some networking settings.
-         */
-        if (mk_list_size(&ins->net_properties) > 0) {
-            ret = flb_config_map_properties_check(ins->p->name,
-                                                  &ins->net_properties,
-                                                  ins->net_config_map);
-            if (ret == -1) {
-                if (config->program_name) {
-                    flb_helper("try the command: %s -o %s -h\n",
-                               config->program_name, ins->p->name);
-                }
-                flb_output_instance_destroy(ins);
-                return -1;
-            }
+
+        /* Init network defaults */
+        flb_net_setup_init(&ins->net_setup);
+
+        if (flb_output_net_property_check(ins, config) == -1) {
+            flb_output_instance_destroy(ins);
+            return -1;
         }
 
         /* Initialize plugin through it 'init callback' */
@@ -1045,7 +1310,13 @@ int flb_output_init_all(struct flb_config *config)
         ret = flb_output_enable_multi_threading(ins, config);
         if (ret == -1) {
             flb_error("[output] could not start thread pool for '%s' plugin",
-                      p->name);
+                      flb_output_name(ins));
+            return -1;
+        }
+
+        /* initialize processors */
+        ret = flb_processor_init(ins->processor);
+        if (ret == -1) {
             return -1;
         }
     }
@@ -1068,6 +1339,18 @@ int flb_output_check(struct flb_config *config)
     return 0;
 }
 
+/* Check output plugin's log level.
+ * Not for core plugins but for Golang plugins.
+ * Golang plugins do not have thread-local flb_worker_ctx information. */
+int flb_output_log_check(struct flb_output_instance *ins, int l)
+{
+    if (ins->log_level < l) {
+        return FLB_FALSE;
+    }
+
+    return FLB_TRUE;
+}
+
 /*
  * Output plugins might have enabled certain features that have not been passed
  * directly to the upstream context. In order to avoid let plugins validate specific
@@ -1098,21 +1381,39 @@ int flb_output_upstream_set(struct flb_upstream *u, struct flb_output_instance *
     if (ins->host.ipv6 == FLB_TRUE) {
         flags |= FLB_IO_IPV6;
     }
+        /* keepalive */
+    if (ins->net_setup.keepalive == FLB_TRUE) {
+        flags |= FLB_IO_TCP_KA;
+    }
 
     /* Set flags */
-    u->flags |= flags;
+    flb_stream_enable_flags(&u->base, flags);
+
+    flb_upstream_set_total_connections_label(u,
+                                             flb_output_name(ins));
+
+    flb_upstream_set_total_connections_gauge(u,
+                                             ins->cmt_upstream_total_connections);
+
+    flb_upstream_set_busy_connections_label(u,
+                                            flb_output_name(ins));
+
+    flb_upstream_set_busy_connections_gauge(u,
+                                            ins->cmt_upstream_busy_connections);
 
     /*
      * If the output plugin flush callbacks will run in multiple threads, enable
      * the thread safe mode for the Upstream context.
      */
     if (ins->tp_workers > 0) {
-        flb_upstream_thread_safe(u);
-        mk_list_add(&u->_head, &ins->upstreams);
+        flb_stream_enable_thread_safety(&u->base);
+
+        mk_list_add(&u->base._head, &ins->upstreams);
     }
 
     /* Set networking options 'net.*' received through instance properties */
-    memcpy(&u->net, &ins->net_setup, sizeof(struct flb_net_setup));
+    memcpy(&u->base.net, &ins->net_setup, sizeof(struct flb_net_setup));
+
     return 0;
 }
 
