@@ -2,8 +2,7 @@
 
 /*  Fluent Bit
  *  ==========
- *  Copyright (C) 2019-2021 The Fluent Bit Authors
- *  Copyright (C) 2015-2018 Treasure Data Inc.
+ *  Copyright (C) 2015-2022 The Fluent Bit Authors
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -26,6 +25,64 @@
 #include <fluent-bit/flb_utils.h>
 #include <fluent-bit/flb_http_server.h>
 
+static struct cmt *metrics_context_create(struct flb_storage_metrics *sm)
+{
+    struct cmt *cmt;
+
+    cmt = cmt_create();
+    if (!cmt) {
+        return NULL;
+    }
+
+    sm->cmt_chunks = cmt_gauge_create(cmt,
+                                      "fluentbit", "storage", "chunks",
+                                      "Total number of chunks in the storage layer.",
+                                      0, (char *[]) { NULL });
+
+    sm->cmt_mem_chunks = cmt_gauge_create(cmt,
+                                          "fluentbit", "storage", "mem_chunks",
+                                          "Total number of memory chunks.",
+                                          0, (char *[]) { NULL });
+
+    sm->cmt_fs_chunks = cmt_gauge_create(cmt,
+                                         "fluentbit", "storage", "fs_chunks",
+                                         "Total number of filesystem chunks.",
+                                         0, (char *[]) { NULL });
+
+    sm->cmt_fs_chunks_up = cmt_gauge_create(cmt,
+                                            "fluentbit", "storage", "fs_chunks_up",
+                                            "Total number of filesystem chunks up in memory.",
+                                            0, (char *[]) { NULL });
+
+    sm->cmt_fs_chunks_down = cmt_gauge_create(cmt,
+                                              "fluentbit", "storage", "fs_chunks_down",
+                                              "Total number of filesystem chunks down.",
+                                              0, (char *[]) { NULL });
+
+    return cmt;
+}
+
+
+/* This function collect the 'global' metrics of the storage layer (cmetrics) */
+int flb_storage_metrics_update(struct flb_config *ctx, struct flb_storage_metrics *sm)
+{
+    uint64_t ts;
+    struct cio_stats st;
+
+    /* Retrieve general stats from the storage layer */
+    cio_stats_get(ctx->cio, &st);
+
+    ts = cfl_time_now();
+
+    cmt_gauge_set(sm->cmt_chunks, ts, st.chunks_total, 0, NULL);
+    cmt_gauge_set(sm->cmt_mem_chunks, ts, st.chunks_mem, 0, NULL);
+    cmt_gauge_set(sm->cmt_fs_chunks, ts, st.chunks_fs, 0, NULL);
+    cmt_gauge_set(sm->cmt_fs_chunks_up, ts, st.chunks_fs_up, 0, NULL);
+    cmt_gauge_set(sm->cmt_fs_chunks_down, ts, st.chunks_fs_down, 0, NULL);
+
+    return 0;
+}
+
 static void metrics_append_general(msgpack_packer *mp_pck,
                                    struct flb_config *ctx,
                                    struct flb_storage_metrics *sm)
@@ -76,29 +133,45 @@ static void metrics_append_input(msgpack_packer *mp_pck,
 {
     int len;
     int ret;
+    uint64_t ts;
     const char *tmp;
     char buf[32];
     ssize_t size;
+    size_t total_chunks;
 
     /* chunks */
     int up;
     int down;
     int busy;
-    int busy_size_err;
+    char *name;
     ssize_t busy_size;
     struct mk_list *head;
     struct mk_list *h_chunks;
     struct flb_input_instance *i;
     struct flb_input_chunk *ic;
 
+    /*
+     * DISCLAIMER: This interface will be deprecated once we extend Chunk I/O
+     * stats per stream.
+     *
+     * For now and to avoid duplication of iterating chunks we are adding the
+     * metrics counting for CMetrics inside the same logic for the old code.
+     */
+
     msgpack_pack_str(mp_pck, 12);
     msgpack_pack_str_body(mp_pck, "input_chunks", 12);
     msgpack_pack_map(mp_pck, mk_list_size(&ctx->inputs));
 
+    /* current time */
+    ts = cfl_time_now();
+
     /* Input Plugins Ingestion */
     mk_list_foreach(head, &ctx->inputs) {
         i = mk_list_entry(head, struct flb_input_instance, _head);
 
+        name = (char *) flb_input_name(i);
+        total_chunks = mk_list_size(&i->chunks);
+
         tmp = flb_input_name(i);
         len = strlen(tmp);
 
@@ -122,6 +195,8 @@ static void metrics_append_input(msgpack_packer *mp_pck,
         msgpack_pack_str(mp_pck, 9);
         msgpack_pack_str_body(mp_pck, "overlimit", 9);
 
+
+        /* CMetrics */
         ret = FLB_FALSE;
         if (i->mem_buf_limit > 0) {
             if (i->mem_chunks_size >= i->mem_buf_limit) {
@@ -129,12 +204,26 @@ static void metrics_append_input(msgpack_packer *mp_pck,
             }
         }
         if (ret == FLB_TRUE) {
+            /* cmetrics */
+            cmt_gauge_set(i->cmt_storage_overlimit, ts, 1,
+                          1, (char *[]) {name});
+
+            /* old code */
             msgpack_pack_true(mp_pck);
         }
         else {
+            /* cmetrics */
+            cmt_gauge_set(i->cmt_storage_overlimit, ts, 0,
+                          1, (char *[]) {name});
+
+            /* old code */
             msgpack_pack_false(mp_pck);
         }
 
+        /* fluentbit_storage_memory_bytes */
+        cmt_gauge_set(i->cmt_storage_memory_bytes, ts, i->mem_chunks_size,
+                      1, (char *[]) {name});
+
         /* status['mem_size'] */
         msgpack_pack_str(mp_pck, 8);
         msgpack_pack_str_body(mp_pck, "mem_size", 8);
@@ -160,6 +249,13 @@ static void metrics_append_input(msgpack_packer *mp_pck,
          * Chunks
          * ======
          */
+
+        /* cmetrics */
+        cmt_gauge_set(i->cmt_storage_chunks, ts, total_chunks,
+                      1, (char *[]) {name});
+
+
+        /* old code */
         msgpack_pack_str(mp_pck, 6);
         msgpack_pack_str_body(mp_pck, "chunks", 6);
 
@@ -169,7 +265,7 @@ static void metrics_append_input(msgpack_packer *mp_pck,
         /* chunks['total_chunks'] */
         msgpack_pack_str(mp_pck, 5);
         msgpack_pack_str_body(mp_pck, "total", 5);
-        msgpack_pack_uint64(mp_pck, mk_list_size(&i->chunks));
+        msgpack_pack_uint64(mp_pck, total_chunks);
 
         /*
          * chunks Details: chunks marked as 'busy' are 'locked' since they are in
@@ -177,7 +273,6 @@ static void metrics_append_input(msgpack_packer *mp_pck,
          */
         busy = 0;
         busy_size = 0;
-        busy_size_err = 0;
 
         /* up/down */
         up = 0;
@@ -192,9 +287,6 @@ static void metrics_append_input(msgpack_packer *mp_pck,
                 if (size >= 0) {
                     busy_size += size;
                 }
-                else {
-                    busy_size_err++;
-                }
             }
 
             if (cio_chunk_is_up(ic->chunk) == CIO_TRUE) {
@@ -206,21 +298,37 @@ static void metrics_append_input(msgpack_packer *mp_pck,
 
         }
 
+        /* fluentbit_storage_chunks_up */
+        cmt_gauge_set(i->cmt_storage_chunks_up, ts, up,
+                      1, (char *[]) {name});
+
         /* chunks['up'] */
         msgpack_pack_str(mp_pck, 2);
         msgpack_pack_str_body(mp_pck, "up", 2);
         msgpack_pack_uint64(mp_pck, up);
 
+        /* fluentbit_storage_chunks_down */
+        cmt_gauge_set(i->cmt_storage_chunks_down, ts, down,
+                      1, (char *[]) {name});
+
         /* chunks['down'] */
         msgpack_pack_str(mp_pck, 4);
         msgpack_pack_str_body(mp_pck, "down", 4);
         msgpack_pack_uint64(mp_pck, down);
 
+        /* fluentbit_storage_chunks_busy */
+        cmt_gauge_set(i->cmt_storage_chunks_busy, ts, busy,
+                      1, (char *[]) {name});
+
         /* chunks['busy'] */
         msgpack_pack_str(mp_pck, 4);
         msgpack_pack_str_body(mp_pck, "busy", 4);
         msgpack_pack_uint64(mp_pck, busy);
 
+        /* fluentbit_storage_chunks_busy_size */
+        cmt_gauge_set(i->cmt_storage_chunks_busy_bytes, ts, busy_size,
+                      1, (char *[]) {name});
+
         /* chunks['busy_size'] */
         msgpack_pack_str(mp_pck, 9);
         msgpack_pack_str_body(mp_pck, "busy_size", 9);
@@ -247,7 +355,7 @@ static void cb_storage_metrics_collect(struct flb_config *ctx, void *data)
     metrics_append_input(&mp_pck, ctx, data);
 
 #ifdef FLB_HAVE_HTTP_SERVER
-    if (ctx->http_server == FLB_TRUE) {
+    if (ctx->http_server == FLB_TRUE && ctx->storage_metrics == FLB_TRUE) {
         flb_hs_push_storage_metrics(ctx->http_ctx, mp_sbuf.data, mp_sbuf.size);
     }
 #endif
@@ -259,16 +367,20 @@ struct flb_storage_metrics *flb_storage_metrics_create(struct flb_config *ctx)
     int ret;
     struct flb_storage_metrics *sm;
 
-    sm = flb_malloc(sizeof(struct flb_storage_metrics));
+    sm = flb_calloc(1, sizeof(struct flb_storage_metrics));
     if (!sm) {
         flb_errno();
         return NULL;
     }
+    sm->cmt = metrics_context_create(sm);
+    if(!sm->cmt) {
+        flb_free(sm);
+        return NULL;
+    }
 
     ret = flb_sched_timer_cb_create(ctx->sched, FLB_SCHED_TIMER_CB_PERM, 5000,
                                     cb_storage_metrics_collect,
-                                    ctx->storage_metrics_ctx,
-                                    NULL);
+                                    ctx->storage_metrics_ctx, NULL);
     if (ret == -1) {
         flb_error("[storage metrics] cannot create timer to collect metrics");
         flb_free(sm);
@@ -326,35 +438,34 @@ static int sort_chunk_cmp(const void *a_arg, const void *b_arg)
 
 static void print_storage_info(struct flb_config *ctx, struct cio_ctx *cio)
 {
+    char *type;
     char *sync;
     char *checksum;
     struct flb_input_instance *in;
 
-    flb_info("[storage] version=%s, initializing...", cio_version());
-
-    if (cio->root_path) {
-        flb_info("[storage] root path '%s'", cio->root_path);
+    if (cio->options.root_path) {
+        type = "memory+filesystem";
     }
     else {
-        flb_info("[storage] in-memory");
+        type = "memory";
     }
 
-    if (cio->flags & CIO_FULL_SYNC) {
+    if (cio->options.flags & CIO_FULL_SYNC) {
         sync = "full";
     }
     else {
         sync = "normal";
     }
 
-    if (cio->flags & CIO_CHECKSUM) {
-        checksum = "enabled";
+    if (cio->options.flags & CIO_CHECKSUM) {
+        checksum = "on";
     }
     else {
-        checksum = "disabled";
+        checksum = "off";
     }
 
-    flb_info("[storage] %s synchronization mode, checksum %s, max_chunks_up=%i",
-             sync, checksum, ctx->storage_max_chunks_up);
+    flb_info("[storage] ver=%s, type=%s, sync=%s, checksum=%s, max_chunks_up=%i",
+             cio_version(), type, sync, checksum, ctx->storage_max_chunks_up);
 
     /* Storage input plugin */
     if (ctx->storage_input_plugin) {
@@ -385,26 +496,36 @@ static int log_cb(struct cio_ctx *ctx, int level, const char *file, int line,
 int flb_storage_input_create(struct cio_ctx *cio,
                              struct flb_input_instance *in)
 {
+    int cio_storage_type;
     struct flb_storage_input *si;
     struct cio_stream *stream;
 
     /* storage config: get stream type */
     if (in->storage_type == -1) {
-        in->storage_type = CIO_STORE_MEM;
+        in->storage_type = FLB_STORAGE_MEM;
     }
 
-    if (in->storage_type == CIO_STORE_FS && cio->root_path == NULL) {
+    if (in->storage_type == FLB_STORAGE_FS && cio->options.root_path == NULL) {
         flb_error("[storage] instance '%s' requested filesystem storage "
                   "but no filesystem path was defined.",
                   flb_input_name(in));
         return -1;
     }
 
+    /*
+     * The input instance can define it owns storage type which is based on some
+     * specific Chunk I/O storage type. We handle the proper initialization here.
+     */
+    cio_storage_type = in->storage_type;
+    if (in->storage_type == FLB_STORAGE_MEMRB) {
+        cio_storage_type = FLB_STORAGE_MEM;
+    }
+
     /* Check for duplicates */
     stream = cio_stream_get(cio, in->name);
     if (!stream) {
         /* create stream for input instance */
-        stream = cio_stream_create(cio, in->name, in->storage_type);
+        stream = cio_stream_create(cio, in->name, cio_storage_type);
         if (!stream) {
             flb_error("[storage] cannot create stream for instance %s",
                       in->name);
@@ -465,28 +586,23 @@ static int storage_contexts_create(struct flb_config *config)
     return c;
 }
 
-static void storage_contexts_destroy(struct flb_config *config)
-{
-    struct mk_list *head;
-    struct flb_input_instance *in;
-
-    /* Iterate each input instance and destroy the context */
-    mk_list_foreach(head, &config->inputs) {
-        in = mk_list_entry(head, struct flb_input_instance, _head);
-        flb_storage_input_destroy(in);
-    }
-}
-
 int flb_storage_create(struct flb_config *ctx)
 {
     int ret;
     int flags;
     struct flb_input_instance *in = NULL;
     struct cio_ctx *cio;
+    struct cio_options opts = {0};
 
     /* always use read/write mode */
     flags = CIO_OPEN;
 
+    /* if explicitly stated any irrecoverably corrupted
+     * chunks will be deleted */
+    if (ctx->storage_del_bad_chunks) {
+        flags |= CIO_DELETE_IRRECOVERABLE;
+    }
+
     /* synchronization mode */
     if (ctx->storage_sync) {
         if (strcasecmp(ctx->storage_sync, "normal") == 0) {
@@ -506,8 +622,14 @@ int flb_storage_create(struct flb_config *ctx)
         flags |= CIO_CHECKSUM;
     }
 
+    /* chunkio options */
+    opts.root_path = ctx->storage_path;
+    opts.flags = flags;
+    opts.log_cb = log_cb;
+    opts.log_level = CIO_LOG_INFO;
+
     /* Create chunkio context */
-    cio = cio_create(ctx->storage_path, log_cb, CIO_LOG_DEBUG, flags);
+    cio = cio_create(&opts);
     if (!cio) {
         flb_error("[storage] error initializing storage engine");
         return -1;
@@ -568,6 +690,7 @@ int flb_storage_create(struct flb_config *ctx)
 void flb_storage_destroy(struct flb_config *ctx)
 {
     struct cio_ctx *cio;
+    struct flb_storage_metrics *sm;
 
     /* Destroy Chunk I/O context */
     cio = (struct cio_ctx *) ctx->cio;
@@ -576,14 +699,13 @@ void flb_storage_destroy(struct flb_config *ctx)
         return;
     }
 
-    if (ctx->storage_metrics == FLB_TRUE &&
-        ctx->storage_metrics_ctx != NULL) {
-        flb_free(ctx->storage_metrics_ctx);
+    sm = ctx->storage_metrics_ctx;
+    if (ctx->storage_metrics == FLB_TRUE && sm != NULL) {
+        cmt_destroy(sm->cmt);
+        flb_free(sm);
+        ctx->storage_metrics_ctx = NULL;
     }
 
     cio_destroy(cio);
-
-    /* Delete references from input instances */
-    storage_contexts_destroy(ctx);
     ctx->cio = NULL;
 }
