@@ -42,6 +42,10 @@ int LLVMFuzzerTestOneInput(const uint8_t *data, size_t size)
     int time_keep = 0;
     int types_len = 0;
 
+    /* Set fuzzer-malloc chance of failure */
+    flb_malloc_mod = 25000;
+    flb_malloc_p = 0;
+
     if (size < 100) {
         return 0;
     }
@@ -148,8 +152,8 @@ int LLVMFuzzerTestOneInput(const uint8_t *data, size_t size)
     flb_utils_print_setup(fuzz_config);
 
     /* now call into the parser */
-    fuzz_parser = flb_parser_create("fuzzer", format, pregex,
-            time_fmt, time_key, time_offset, time_keep, 0,
+    fuzz_parser = flb_parser_create("fuzzer", format, pregex, FLB_TRUE,
+            time_fmt, time_key, time_offset, time_keep, 0, FLB_FALSE,
             types, types_len, list, fuzz_config);
 
     /* Second step is to use the random parser to parse random input */
